/************************************************************************
  				mutator.h - SBMLEvolver - CSB, FSU Jena

This software is free for use and modification, but comes without any
warranty. Usage is entirely at your own risk!

Developers: 	Thorsten Lenser (thlenser@inf.uni-jena.de)


This class provides the building blocks from which the EA can choose 
the components of the model. These are:
- a list of possible species
- a tree of possible rate rules


	$Id: buildingblocks.h 3 2007-06-04 12:59:46Z thlenser $


**************************************************************************/


#ifndef BUILDINGBLOCKS_H_
#define BUILDINGBLOCKS_H_

#include <string>
#include <vector>
#include <sstream>

#include "include.h"
#include "sbmltools.h"

using namespace std;

namespace sbmlevolver
{

/**
 * \brief Manages building blocks
 *
 * This class provides the building blocks from which the EA can choose 
 * the components of the model. These are:
 *	- a list of possible species
 *	- a tree of possible rate rules
 */
class BuildingBlocks
{
public:
	/// Default constructor, does nothing
	BuildingBlocks();
	
	/// Construct object from a text string 
	BuildingBlocks(string BBstring);
	virtual ~BuildingBlocks();
	
	/// Learn new reaction types from a SBML model	
	void learnReactionsFromModel(Model*);
	
	/// Return the minimal number of species in the model
	int getMinNumSpecies();
	
	/// Return the maximal number of species in the model
	int getMaxNumSpecies();
	
	/// Return the minimal number of reactions in the model
	int getMinNumReactions();
	
	/// Return the maximal number of reactions in the model
	int getMaxNumReactions();
	
	/// Return species with the given number or NULL if no such species exists
	Species* getSpecies(int);
	
	/// Return reaction with the given characteristics or NULL if no such reaction exists
	Reaction* getReaction(int,int,int,int);
	
	/// Return reaction with the given number when listing reactions linearly
	Reaction* getReaction(int);
	
	/// Return reaction id with the given number
	string getReactionId(int);
	
	/// Return the first species id that is not already present in the given model
	Species* getSuitableSpecies(Model*);
	
	/// Return the first reaction id that is not already present in the given model
	string getSuitableReactionId(Model*);
	
	/// Return number of rate rules with given number of reactants, products and modifiers involved
	int getNumReactions(int,int,int);
	
	/// Return the absolute number of stored reactions
	int getNumReactions();
	
	/// Return the type-number of the given reaction
	int getReactionType(Reaction*);
	
	/// Return the specified rate rule with the given number of reactants, products and modifiers involved, or NULL if no such thing exists
	KineticLaw* getKineticLaw(int,int,int,int);
	
	/// Indicate whether the specified reaction type is reversible or not
	bool isReversible(int,int,int,int);

	int getNumTransportReactions();
	int getNumNonTransportReactions();
	Reaction* getTransportReaction(int);
	Reaction* getNonTransportReaction(int);
	
	/// Print the object to an output stream (for debugging)
	void print(ostream* out);
	
	/// Get the lowest initial species concentration allowed
	string getLowestInitialConcentration();
	
	/// Get the highest initial species concentration allowed
	string getHighestInitialConcentration();
	
	
private:
	/// Minimal number of species
	int minNumSpecies;
	/// Maximal number of species
	int maxNumSpecies;
	
	/// Minimal number of reactions
	int minNumReactions;
	/// Maximal number of reactions
	int maxNumReactions;
	
	/// Number of stored reactions
	int numStoredReactions;
	
	/// Stored reaction names
	vector<Species*> species;
	
	/// lower bound for species initial concentration
	string speciesInitLowerBound;
	
	/// upper bound for species initial concentration
	string speciesInitUpperBound;

	/// indices of transport reactions
	vector<int> transport_reactions;
	bool index_transport_reactions_exists;

        /// indices of non-transport reactions
	vector<int> non_transport_reactions;
	bool index_non_transport_reactions_exists;
	
	/// Return generic species id with the given number ("X<num>")
	string getSpeciesId(int);

	/// get the type ID of the reaction signature
	int getTypeName(int numRe, int numPr, int numMo, int num);
   	
   	/// Contains the possible reactions to be used in the model. Indexes are "number of reactants", "number of products", "number of modifiers", and "reaction number", in this order.
   	vector< vector< vector< vector< Reaction* > > > > reactions;
   	
   	/// Add a new reaction type (kinetic law)
   	void addReactionTypeAsBuildingBlock(KineticLaw* kl, int numReactants, int numProducts, int numModifiers);
	
	/// Add a specific reaction as a building block
	void addReactionAsBuildingBlock(Reaction* re);
};

} // namespace sbmlevolver

#endif /*BUILDINGBLOCKS_H_*/
