/* Version information
 *
 * $Id: centralcontrol.cpp 29 2007-06-15 13:35:45Z thlenser $
 *
 */

#include "centralcontrol.h"


#define WORKTAG	1
#define DIETAG	2
#define RESULTTAG	3
#define MASTER	0
#define READY_FOR_MIGRATION "ready_for_migration"

using namespace std;
using namespace sbmlevolver;


/* CentralControl::CentralControl() {
	params = new ParamObj();
	
	currentTurn = 0;
	startTime = time(NULL);
	bestFitness = VERY_BAD_FITNESS;
} */

CentralControl::CentralControl(int argc, char *argv[]) {	
	// Check if only the fitness of one model is to be calculated
	if((argc > 1) && (string(argv[1]) == string("--get-fitness"))){
		cout << calculateOnlyFitness(argc,argv) << endl;
		exit(0);
	}

	// Check if only the parameters of one model are to be fitted
	if((argc > 1) && (string(argv[1]) == string("--fit-parameters"))){
		fitOnlyParameters(argc,argv);
		exit(0);
	}

	// Check if we want to run a mutation study 
	if((argc > 1) && (string(argv[1]) == string("--study-mutations"))){
		studyMutations(argc,argv);
		exit(0);
	}	

	// Check if help is to be printed
	if((argc > 1) && ((string(argv[1]) == string("--help")) || (string(argv[1]) == string("-h")))){
		params = new ParamObj();
		params->printHelp(cout);
		delete params;
		exit(0);
	}

	// check if name of optionsfile is given
	string optionsfile;
	if(argc > 1) optionsfile = string(argv[1]);
	else optionsfile = string("options.txt");
	stringstream more_options;
	for(int i=2; i < argc; i++){
		more_options << argv[i];
		more_options << " ";
	}
	params = new ParamObj(optionsfile);
	params->parseOptions((istream*) &more_options);
	
	currentTurn = 0;
	startTime = time(NULL);
	bestFitness = VERY_BAD_FITNESS;

	ofstream call_log("call.log");
	params->print(call_log);
	call_log.close();
}

void CentralControl::Run() {
	// Initialise rng: Seed depends on hostname, time and pid.
	rng = new RandomNumberGenerator(params->seed);
        #ifndef USE_MPI_VERSION
	RunOnSingleProcessor();
	#endif
	
	#ifdef USE_MPI_VERSION
	MPI::Init();
	
	int myrank = MPI::COMM_WORLD.Get_rank();
	
	srand( myrank );
	
	if( myrank == MASTER ) {
		RunMaster();
	} else if((myrank > MASTER) && (myrank <= MASTER + params->population.num_populations)){
			RunPopulation();
	}else{
		RunSlave();	
	}
	
	MPI::Finalize();
	return;
	#endif
}

#ifndef USE_MPI_VERSION
void CentralControl::RunOnSingleProcessor(){

	set<Individual*> *inds;
	set<Individual*>::iterator ind_iter;

	ofstream outputFile;
	// init workspace: set CWD to output directory
	if (!(params->session.outputDir.empty()) && chdir(params->session.outputDir.c_str()) != 0){
		mkdir(params->session.outputDir.c_str(),S_IRWXU);
		if(chdir(params->session.outputDir.c_str()) != 0){
			THROW_ERROR("could not change to output directory");
		}
	}
	outputFile.open(params->session.outputFile.c_str());

	ObjectiveFunction* of = new ObjectiveFunction(params->OFstring, rng->getASeed());

	ParameterFitter* paraFit = new ParameterFitter(params->paraFitter.populationSize,params->paraFitter.numOfParents,params->paraFitter.numOfGenerations,params->paraFitter.useOldStartingValues,rng->getASeed());

	if(params->population.num_populations > 1) cout << "Pop\t ";
	cout << "Turn\t Best\t\t Avg\t\t Diff1\t\t Diff2\t\t Species\t Reactions" << endl;
	
	vector<Population*> pop;
	Individual* ind;

	// init population
 	for(int i=0; i < params->population.num_populations; i++){
		pop.push_back(new Population(rng->getASeed(), params));
		
		// Rate the whole population initially
		pop.at(i)->unsetFitnessOfWholePopulation();
		// calculate the fitnesses
		ind = pop.at(i)->extractTask();
		while (ind != NULL) {
			// calc fitness
			ind->calculateFitness(paraFit,of); // Force fitness to be recalculated, even if it is set

			pop.at(i)->submitResult(ind);

			ind = pop.at(i)->extractTask();
		}
 	}
		
	
	int count = 0;
	int count_save = 1;
	bool stop_next_turn = false;
	do {
		currentTurn++;
		if(stop_next_turn) stopRun = true;

		// proceed one generation
		for(int i=0; i < params->population.num_populations; i++){
			/* "init" new generation */
			pop.at(i)->incCurrentGeneration();
			
			/* generate offspring */
			pop.at(i)->createOffspring();
		
			/* check if we should re-evaluate the population */
			if( params->population.alwaysEvaluate ) pop.at(i)->unsetFitnessOfWholePopulation();	
				
			/* calculate the fitnesses */
			ind = pop.at(i)->extractTask();
			while (ind != NULL) {
				// calc fitness
				ind->calculateFitness(paraFit,of); // Force fitness to be recalculated, even if it is set
		
				pop.at(i)->submitResult(ind);

				ind = pop.at(i)->extractTask();
			}

			/* adapt fitnesses using kernel-based fitness sharing */
			//pop->fitnessSharing();
			
			/* remove the unfittest */
			pop.at(i)->select();
			
			if(pop.at(i)->getBestFitness() < bestFitness) bestFitness = pop.at(i)->getBestFitness();

			// Maybe save some individuals?
			if((count_save == params->session.saveEvery) && (!stopRun)){
				inds = pop.at(i)->getBestIndividuals(params->session.numSave);
				ind_iter = inds->begin();
				while (ind_iter != inds->end()) {
					Individual* save_ind = new Individual(*ind_iter); // copy it to avoid side-effects when readjusting species names
					save_ind->WriteToFile();
					delete save_ind;
					ind_iter++;
				}
				delete inds;
				count_save = 0;
			}

			string output = generateOutput(pop.at(i));
			if(params->population.num_populations > 1) cout << i << " ";
			cout << output << flush;
			if(params->population.num_populations > 1) outputFile << i << " ";
			outputFile << output << flush;
			if(checkTermination()) stopRun = true;
		}
			
		// Check for migration. Migrate in a ring-like fashion.
		count++;
		if((params->population.num_populations > 1) && (count > params->population.isolationTime) && (currentTurn > 1)){ // no migration in first round when no fitness is available
			for(int i=0; i < params->population.num_populations; i++){
				int destinationPop = i-1;
				if(destinationPop == -1) destinationPop = params->population.num_populations -1;
				// If you are number 1, send first, all others receive first
				Individual* bestInd = pop.at(i)->extractBestIndividual();
				pop.at(destinationPop)->receiveIndividual(bestInd);
				count = 0;					
			}
		}

		count_save++;
		
	} while (!stopRun);

	/*** safe results & tidy up ***/
	// safe population
	for(int i=0; i < params->population.num_populations; i++){
		inds = pop.at(i)->getBestIndividuals(params->population.mu);
		ind_iter = inds->begin();
		while (ind_iter != inds->end()) {
			(*ind_iter)->WriteToFile(); // the Ind determines the correct filename itself
			ind_iter++;
		}
		delete inds;
		delete pop.at(i);
	}
	delete params;
	delete rng;
	delete of;
	delete paraFit;
	outputFile.close();
}
#endif

#ifdef USE_MPI_VERSION
void CentralControl::RunSlave() {
    int myrank = MPI::COMM_WORLD.Get_rank();

	ObjectiveFunction* of = new ObjectiveFunction(params->OFstring, rng->getASeed());
	ParameterFitter* paraFit = new ParameterFitter(params->paraFitter.populationSize, params->paraFitter.numOfParents, params->paraFitter.numOfGenerations, rng->getASeed());
	
	Individual* ind;
	
	//cout << myrank << ": Slave starting" << endl;
	
	MPI::Status status;
	
	int length;
	
	int my_pop;
	//my_pop = MASTER;
	// calculate who your master is
	int num_pop = params->population.num_populations;
	int nslaves = MPI::COMM_WORLD.Get_size() - 1 - params->population.num_populations;
	int num_workers = nslaves / num_pop;
	my_pop = (myrank - 1 - num_pop) / num_workers +1;

	// Are you currently working?
	bool currentlyWorking = false;

	/* main loop */
	do {
		currentTurn++;

		if(!currentlyWorking){
			// Tell the master you are ready to work
			sendString(my_pop, string("ready"));
		}
			
		/* receive Ind */
		string ind_str = recvString(&my_pop);
		if(ind_str == string("stop")){
			stopRun = true;
		}
		else if(ind_str == string("no_work")){
			currentlyWorking = false;
			//cout << myrank << "; Idle " << currentTurn << endl;
		}
		else{
			currentlyWorking = true;
			//cout << myrank << ": Received task" << endl;	
			ind = new Individual(ind_str);
			
			/* calculate fitness */
			ind->calculateFitness(paraFit,of);
			if (isnan(ind->getFitness())) {
				THROW_WARN("WARN: NaN fitness detected");
				ind->setFitness(VERY_BAD_FITNESS +1);
			}
	
			/* send serialized int */
			ind_str = ind->Serialize();
			sendString(my_pop,ind_str);
			delete ind;
	
			/* output */
			//cout << myrank << ": Finished " << currentTurn << endl;
		}
	} while (!stopRun);

	delete of;
	delete paraFit;
}
#endif

#ifdef USE_MPI_VERSION
void CentralControl::RunPopulation(){
	set<Individual*> *inds;
	set<Individual*>::iterator ind_iter;
	
	int num_populations = params->population.num_populations;
	
	// init workspace: set CWD to input directory
	if (!(params->session.inputDir.empty()) && chdir(params->session.inputDir.c_str()) != 0) THROW_ERROR("could not change to input directory");

	Population* pop;
	Individual* ind;

	// init population
 
	pop = new Population(rng->getASeed(), params);
	
	// Determine my workers
	int myrank = MPI::COMM_WORLD.Get_rank();
	int num_pop = params->population.num_populations;
	int nslaves = MPI::COMM_WORLD.Get_size() - 1 - params->population.num_populations;
	// Master is 0, pops are 1..num_pop, workers are num_pop+1..num_pop+nslaves
	// determine number of slaves per population
	if(nslaves % num_pop != 0) THROW_ERROR("Number of workers is not a multiple of the number of populations!");
	int num_workers = nslaves / num_pop;
	vector<int> my_workers;
	for(int i = (num_pop+1) + (myrank-1)*num_workers; i < (num_pop+1) + (myrank)*num_workers; i++){
		my_workers.push_back(i);
	}		
	
	// Rate the whole population initially
	pop->unsetFitnessOfWholePopulation();

	int count = 0;
	int count_save = 1;
	bool stop_next_turn = false;
	//MPI::Request req1, req2;
	do {
		currentTurn++;
		if(stop_next_turn) stopRun = true;

		// proceed one generation
		// migration every x generations (before pop so that new ones are subject to selection)
		if((num_populations > 1) && (!stopRun)){
			count++;
			if((count > params->population.isolationTime) && (currentTurn > 1)){ // no migration in first round when no fitness is available
				// Tell the master you're ready for migration 
				// synchronisation is needed here anyway
				int master = MASTER;
				string readystr = string(READY_FOR_MIGRATION);
				sendString(master,readystr);
				// Wait for answer, when it comes all populations are ready
				string answer = recvString(&master);
				if(answer == string("stop")){
					//cout << "stopping pop " << myrank << endl;
					stopRun = true;
				}
				else{
					if(answer != string("migrate")){
						THROW_ERROR("Received strange response from master control process");
					}
					
					int destinationPop;
					//destinationPop = rng->uniform(1,num_populations);
					//while(destinationPop == myrank) destinationPop = rng->uniform(1,num_populations);
					destinationPop = (myrank -1); // ring topology
					if(destinationPop == 0) destinationPop = num_pop;
					// If you are number 1, send first, all others receive first
					Individual* bestInd = pop->extractBestIndividual(); // Get it here already to avoid it being the received ind
					Individual* recvInd;
					string indstr;
					if(myrank == 1){
						indstr = bestInd->Serialize(); 
						//cout << "pop " << myrank << "sends " << indstr << endl;
						sendString(destinationPop,indstr);
						delete bestInd;
					}
					// In a ring, we can only receive from myrank+1
					int source = myrank + 1;
					if(source == (num_pop +1)) source = 1;
					indstr = recvString(&source);
					//cout << "pop " << myrank << "received " << indstr << endl;
					recvInd = new Individual(indstr);
					pop->receiveIndividual(recvInd);

					//cout << "pop " << myrank << ": received individual from pop " << source << endl;
					if(myrank != 1){
						indstr = bestInd->Serialize();
						//cout << "pop " << myrank << "sends " << indstr << endl;
						sendString(destinationPop,indstr);
						delete bestInd;
					}
					count = 0;					
				}
			}
		}
		
		/* "init" new generation */
		pop->incCurrentGeneration();	
			
		/* generate offspring */
		pop->createOffspring();
		
		/* check if we should re-evaluate the population */
		if( params->population.alwaysEvaluate ) pop->unsetFitnessOfWholePopulation();	
				
		/* calculate the fitnesses */
		// Wait for "ready" message or a result from a client, then send him new work
		ind = pop->extractTask();
		while((ind != NULL) || (pop->tasksInProgress() > 0)){
			int source = MPI::ANY_SOURCE;
			
			string ind_string = recvString(&source);
			if(ind_string == string("stop")){// This would come from the server
				stopRun = true;
				continue;
			}
			
			if(ind_string != string("ready")){
				Individual* finishedInd = new Individual(ind_string);
				pop->submitResult(finishedInd);
				finishedInd = NULL;
			}
			
			//cout << "pop " << myrank << ": I received results from slave #" << source << endl;
		       
			if(ind != NULL){	// If null then no task is left
				//cout << "pop " << myrank << ": Sending task to slave #" << source << endl;
				ind_string = ind->Serialize();
				sendString(source, ind_string);
				delete ind;
			}
			else{
				// If run is ended, tell client to stop
				if(stopRun){
					sendString(source,string("stop"));
				}
				// Tell client that there is no more work
				else sendString(source,string("no_work"));
			}

			//cout << "pop " << myrank << ": pop size " << pop->getNumOfIndividuals() << endl; 
			ind = pop->extractTask();

		}

		/* adapt fitnesses using kernel-based fitness sharing */
		//pop->fitnessSharing();
			
		/* remove the unfittest */
		pop->select();
			
		if(pop->getBestFitness() < bestFitness) bestFitness = pop->getBestFitness();
		double currGen = (double) pop->getCurrentGeneration();
		mean_best_fit_over_all_generations = (currGen-1) / currGen * mean_best_fit_over_all_generations + 1.0 / currGen * pop->getBestFitness();

		// Maybe save some individuals?
		if((count_save == params->session.saveEvery) && (!stopRun)){
			inds = pop->getBestIndividuals(params->session.numSave);
			ind_iter = inds->begin();
			while (ind_iter != inds->end()) {
				Individual* save_ind = new Individual(*ind_iter); // copy it to avoid side-effects
				// TODO: save results of populations to different subdirectories, to distinguish models
				save_ind->WriteToFile();
				delete save_ind;
				ind_iter++;
			}
			delete inds;
			count_save = 0;
		}

		// Now send the output to the master controller
		// If the run is over, do not send any more results to the master
		if(!stopRun){
			string output = generateOutput(pop);
			int master = MASTER;
			sendString(master,output);
			// Receive "go_on" or "stop" signal
			string signal = recvString(&master);
			if(signal == string("stop")){
				stop_next_turn = true;
			}	
		}


		count_save++;
		
	} while (!stopRun);

	/*** safe results & tidy up ***/
	// safe population
	inds = pop->getBestIndividuals(params->population.mu);
	ind_iter = inds->begin();
	while (ind_iter != inds->end()) {
		(*ind_iter)->WriteToFile(); // the Ind determines the correct filename itself
		ind_iter++;
	}
	delete inds;
	delete pop;
	delete params;
	delete rng;
}
#endif


#ifdef USE_MPI_VERSION
void CentralControl::RunMaster() {
	ofstream outputFile;

	// init workspace: set CWD to output directory
	if (!(params->session.outputDir.empty()) && chdir(params->session.outputDir.c_str()) != 0) THROW_ERROR("could not change to output directory");
	outputFile.open(params->session.outputFile.c_str());

	// update screen & output files, wait for migration signals
	int num_pops = params->population.num_populations;
	int source = MPI::ANY_SOURCE;
	set<int> migration_ready_pops;
	do{
		string output = recvString(&source);

		if(output == string(READY_FOR_MIGRATION)){
			migration_ready_pops.insert(source);
			if(migration_ready_pops.size() == num_pops){
				//cout << "Migrate!" << endl;
				string migrate("migrate");
				for(int i=1; i <= num_pops; i++){
					sendString(i,migrate);
				}
				migration_ready_pops.clear();
			}
		}
		else{
			currentTurn++;
		      	cout << "pop " << source << ": " << output;
			outputFile << output;
			// Give feedback 
			string go_on("go_on");
			sendString(source, go_on);
		}
		source = MPI::ANY_SOURCE; // source is changed in recvString
	}while(!checkTermination());
	
	// Send stop signal to all populations
	string stop("stop");
	for(int i=1; i <= num_pops; i++){
		// Check if population is waiting for reply, otherwise we still need to receive one message
		if(migration_ready_pops.find(i) == migration_ready_pops.end()){
			string output = recvString(&i);
			
			if(output != string(READY_FOR_MIGRATION)){
				cout << "pop " << i << ": " << output << "\t" << i;
				outputFile << output << "\t" << i;
			}
		}
		//cout << "sending stop signal to pop " <<  i << endl;
		sendString(i,stop);
	}
	
	delete params;
	delete rng;

	outputFile.close();
}
#endif

string CentralControl::generateOutput(Population* pop) {
	stringstream tmp;
	tmp.setf(ios::fixed, ios::floatfield);
	tmp.setf(ios::showpoint);
	//tmp.precision(5);
	tmp.width(5);
	tmp << pop->getCurrentGeneration() << "\t" 
	        << pop->getBestFitness() << "\t"
	        << pop->getAverageFitness() << "\t"
		    << pop->getAverageStructuralDifference() << "\t"
		<< pop->getAverageStructuralDifference2() << "\t"
		<< pop->getAverageNumSpecies() << "\t"
	        << pop->getAverageNumReactions() << endl;

	return tmp.str();
}


bool CentralControl::checkTermination() {
	char c;
	
	if(stopRun){
		cout << "Run finished: received signal to stop" << endl;
		return true; // Signal SIGUSR1 has been caught
	}
	
	// number of iterations
	if (params->session.stopTurn && (currentTurn >= params->session.stopTurn)) {
		cout << "Run finished: maximal number of turns reached" << endl;
		return true;
	}
	
	// time
	if (params->session.stopTime && (time(NULL) >= startTime + params->session.stopTime)) {
 		cout << "Run finished: maximal calculation time exceeded" << endl;
		return true;
	}
	
	// fitness
	if (bestFitness <= params->session.stopFit) {
 		cout << "Run finished: reached desired fitness" << endl;
		return true;
	}
		
	// nothing met ? -> let's continue
	return false;
}




// Catch the signal SIGUSR1 and set the flag CentralControl::stopRun to true
void CentralControl::catchStopSignal(int sig){
	CentralControl::stopRun = true;
}

bool CentralControl::stopRun = false; // used to stop the run when signal SIGUSR1 is received

int main (int argc, char *argv[]) {	
	// register signal handler
	signal(SIGUSR1, CentralControl::catchStopSignal);
	
	CentralControl *c;
  
	c = new CentralControl(argc, argv);
  
	c->Run();
  
  	delete c;
  	
  	return 0;
}



// Helper function to use MPI to send a string
#ifdef USE_MPI_VERSION
void CentralControl::sendString(int dest, string message){
	int length = message.size() +1;
	const char* tmp_char = message.c_str();
			
	MPI::COMM_WORLD.Send( &length, 1, MPI::INT, dest, WORKTAG );
	MPI::COMM_WORLD.Send( tmp_char, length, MPI::CHAR, dest, WORKTAG );	
}
#endif

#ifdef USE_MPI_VERSION
void CentralControl::sendStringNonblocking(int dest, string message, MPI::Request* req1, MPI::Request* req2){
	int length = message.size() +1;
	const char* tmp_char = message.c_str();
			
	*req1 = MPI::COMM_WORLD.Isend( &length, 1, MPI::INT, dest, WORKTAG);
	*req2 = MPI::COMM_WORLD.Isend( tmp_char, length, MPI::CHAR, dest, WORKTAG);	
}
#endif

// Helper function to receive a string via MPI (from any source)
#ifdef USE_MPI_VERSION
string CentralControl::recvString(int* source){
	MPI::Status status;
	int length;
	MPI::COMM_WORLD.Recv( &length, 1, MPI::INT, *source, MPI::ANY_TAG, status );
	*source = status.Get_source();
	char tmp_char[length];			
	MPI::COMM_WORLD.Recv( tmp_char, length, MPI::CHAR, *source, MPI::ANY_TAG, status );
	string tmp_string = string(tmp_char);
	return tmp_string;
}
#endif




/************************
 *
 * Special routine to calculate the fitness of just one model.
 * Used when --get-fitness is the first option.
 * 
 ***********************/

double CentralControl::calculateOnlyFitness(int argc, char *argv[]){
	// Get model name
	char* model_filename;
	if(argc < 3){
		cerr << "No name of model given!" << endl;
		exit(-1);
	}
	else model_filename = argv[2];
	if(argc > 3){
		cerr << "Too many arguments for --get-fitness option" << endl;
		exit(-1);
	}
	
	
	// Defaults
	unsigned long int seed = (unsigned long int) time(NULL) + getpid();
	
	Model* m;
	ObjectiveFunction* F;
	
	F = new ObjectiveFunction(seed); // load with default file "objectivefunction.txt"
	F->setStoreRun(true);
	m = LoadSBMLModel(model_filename);

	double fitness = F->evaluate(m);
	
	delete F;
	delete m;
	
	return fitness;
}



/********************************************************************
 * 
 * Extension to use the parameter fitting es only
 *
 ********************************************************************/
void CentralControl::fitOnlyParameters(int argc, char *argv[]){
	// Get model name
	char* model_filename;
	if(argc < 3){
		THROW_ERROR("Parameter fitter called but no filename of model is given!");
	}
	else model_filename = argv[2];
	
	int mu, lambda, numGenerations;
	bool quiet = false;
	bool show_last_result = false;
	
	unsigned long int seed;
	//defaults
	mu = 30;
	lambda = 60;
	numGenerations = 100;
	seed = (int) time(NULL) + getpid();
	quiet = false;
	show_last_result = false;

	string objective_filename;
	
	// read arguments
	for(int counter=3; counter < argc; counter++){
		if(string(argv[counter]) == string("--mu")){
			counter ++;
			mu = atoi(argv[counter]);
		}
		else if(string(argv[counter]) == string("--lambda")){
			counter++;
			lambda = atoi(argv[counter]);
		}
		else if(string(argv[counter]) == string("--num-gen")){
			counter++;
			numGenerations = atoi(argv[counter]);
		}
		else if(string(argv[counter]) == string("--seed")){
			counter++;
			seed = atoi(argv[counter]);
		}
		else if(string(argv[counter]) == string("--objective")){
		  counter++;
		  objective_filename = string(argv[counter]);
		}
		else if(string(argv[counter]) == string("--quiet")) quiet = true;
		else if(string(argv[counter]) == string("--last-result")) show_last_result = true;
		else{
			cout << "Unrecognized command line option: " << argv[counter] << endl;
			exit(-1);
		}
	}

	if(objective_filename.empty()) objective_filename = string("objectivefunction.txt");
	
	Model* m;
	ObjectiveFunction* F;
	ParameterFitter* paraFit;
	
	cout << "Loading " << objective_filename << " now" << endl;
	F = new ObjectiveFunction(seed + 1001,objective_filename.c_str()); // load with default file "objectivefunction.txt"
	cout << "Loaded " << objective_filename << " as objective function"  << endl;
	paraFit = new ParameterFitter(lambda, mu, numGenerations, false, seed);
	m = LoadSBMLModel(model_filename);
	
	// Set boundaries to all unbounded parameters
	Parameter* par;
	Reaction* re;
	Species* sp;

	for(int i=0; i < m->getNumParameters(); i++){
		par = m->getParameter(i);
		if(!isSetAttr(par,"lower_bound")) setAttr(par, "lower_bound", 0);
		if(!isSetAttr(par,"upper_bound")) setAttr(par, "upper_bound", LARGEST_PARAMETER_STRING);
	}
	for(int i=0; i < m->getNumReactions(); i++){
		re = m->getReaction(i);
		for(int j=0; j < re->getKineticLaw()->getNumParameters(); j++){
			par = re->getKineticLaw()->getParameter(j);
			if(!isSetAttr(par,"lower_bound")) setAttr(par,"lower_bound",0);
			if(!isSetAttr(par,"upper_bound")) setAttr(par,"upper_bound",LARGEST_PARAMETER_STRING);

			// Just for gene_reg
			  /*
			if(par->getId() == string("b")){
			  if(!isSetAttr(par,"lower_bound")) setAttr(par,"lower_bound",0);
			  if(!isSetAttr(par,"upper_bound")) setAttr(par,"upper_bound",1);

			}
			else if((par->getId() == string("k")) || (par->getId() == string("k1")) || (par->getId() == string("k2"))){
			  if(!isSetAttr(par,"lower_bound")) setAttr(par,"lower_bound",0);
			  if(!isSetAttr(par,"upper_bound")) setAttr(par,"upper_bound",5);
			}
			else if((par->getId() == string("Kd")) || (par->getId() == string("Kd1")) || (par->getId() == string("Kd2"))){
			  if(!isSetAttr(par,"lower_bound")) setAttr(par,"lower_bound",0);
			  if(!isSetAttr(par,"upper_bound")) setAttr(par,"upper_bound",10);
			}
			else if((par->getId() == string("h")) || (par->getId() == string("h1")) || (par->getId() == string("h2"))){
			  if(!isSetAttr(par,"lower_bound")) setAttr(par,"lower_bound",0);
			  if(!isSetAttr(par,"upper_bound")) setAttr(par,"upper_bound",3);
			}
			  */
		}
	}	
	par = NULL;
	for(int i=0; i < m->getNumSpecies(); i++){
		sp = m->getSpecies(i);
		if(!isSetAttr(sp,"lower_bound")) setAttr(sp, "lower_bound", 0);
		if(!isSetAttr(sp,"upper_bound")) setAttr(sp, "upper_bound", LARGEST_INITIAL_CONCENTRATION_STRING);
		//if(!isSetAttr(sp,"upper_bound")) setAttr(sp, "upper_bound", 10);
	}		
	
	Model* fitted_m = (Model*) m->clone();
	
	paraFit->fitParameters(fitted_m, F,quiet);
	
	//if(show_last_result) cout << paraFit->best_fitness << endl;
	
	char fitted_model_file[1000];
	sprintf(fitted_model_file, "fitted_%s", model_filename);
	cout << "Saving " << fitted_model_file << endl;	
	SaveSBMLModel(fitted_m, fitted_model_file);
	cout << "Done" << endl;
	
	delete F;
	delete paraFit;
	delete m;
	delete fitted_m;
}




/********************************************************************
 * 
 * Extension to study the effect of mutations on behaviour
 * TODO: delete this before release!!!
 * TODO: also delete include directives for buildingblocks, mutator and integrator
 *
 ********************************************************************/
void CentralControl::studyMutations(int argc, char *argv[]){
	// Get model name
	char* model_filename;
	if(argc < 3){
		THROW_ERROR("Mutation study called but no filename of model is given!");
	}
	else model_filename = argv[2];
	
	int typeOfMutation;
      
	// read arguments
	if(argc < 4){
		THROW_ERROR("Not enough arguments given: type of mutation missing!");
	}
	else if(argc == 4){
		typeOfMutation = atoi(argv[3]);
	}
	else if(argc > 4){
		THROW_ERROR("Too many arguments!");
	}
	
	Model *m1, *m2;
	m1 = LoadSBMLModel(model_filename);
	m2 = (Model*) m1->clone();

	vector<double> mutProbs(7,0.0);
	mutProbs.at(typeOfMutation) = 1.0;
	// meanings:
	// 0: nothing
	// 1: add species
	// 2: delete species
	// 3: add reaction
	// 4: delete reaction
	// 5: replace reaction
	// 6: duplicate species

	string BBstring;
	ifstream input;
	char tmp_char;
	input.open("buildingblocks.txt");
	if(input){	// the file is there
		while(! input.eof()){
			input.get(tmp_char);
			BBstring.push_back(tmp_char);
		}
	}
	input.close();
	BuildingBlocks* bb = new BuildingBlocks(BBstring);

	Mutator* mut = 	new Mutator((int) time(NULL) + getpid(), bb, mutProbs);
	
	// create mutation of m1
	mut->mutateModel(m2);
	
	char new_model_file[1000];
	sprintf(new_model_file, "mutated_%d_%s", typeOfMutation, model_filename);
	//cout << "Saving " << fitted_model_file << endl;	
	SaveSBMLModel(m2, new_model_file);
	//cout << "Done" << endl;



	//cout << "Number of species in original model: " << m1->getNumSpecies() << endl;
	//cout << "Number of species in mutated model: " << m2->getNumSpecies() << endl;
	//cout << "Number of reactions in original model: " << m1->getNumReactions() << endl;
	//cout << "Number of reactions in mutated model: " << m2->getNumReactions() << endl;
	vector<double> x = behaviouralDifference(m1,m2);
	cout << "Behavioural difference D1: ";
	cout <<  x.at(0) << endl;
	cout << "Behavioural difference D2: ";
	cout <<  x.at(1) << endl;
	
	delete mut;
	delete bb;
	delete m1;
	delete m2;
}


vector<double> CentralControl::behaviouralDifference(Model* m1, Model* m2){
		
	// getting cvode settings from the objective function
	int printstep = 99;
	double time = 100;
	
	integrator_result* results1;
	integrator_result* results2;

	Integrator* integrator = new Integrator();

	results1 = integrator->integrateSOSlib(m1, time, printstep, false); 
	
	if(results1 == NULL){
		cout << "Could not solve ODE for model m1! ERROR1" << endl;
		exit(0);
		//THROW_WARN("Error when integrating model m1! Returning large difference...");
		//return 10000;
	}
	
	results2 = integrator->integrateSOSlib(m2, time, printstep, false); 
	if(results2 == NULL){
		if(results1) delete results1;
		cout << "Could not solve ODE for model m2! ERROR2" << endl;
		exit(0);
	}
	
	double diff1 = 0;
	double diff2 = 0;
	double tmp = 0;
	int num_common_species = 0;
	map< string, vector<double> >::iterator iter;
	for(int i=0; i < results1->ids.size(); i++){	
		iter = results2->values.find(results1->ids.at(i));
		if(iter == results2->values.end()){
			continue;
		}
		else{
			vector<double> x2 = iter->second;
			num_common_species++;
			vector<double> x1 = results1->values[results1->ids.at(i)];
			tmp = 0;
			for(int k=0; k < x1.size(); k++){
				diff1 += pow(x1.at(k) - x2.at(k), 2);
				tmp += pow(x1.at(k) - x2.at(k), 2);
			}			
			if(tmp > diff2) diff2 = tmp;
		}		
	}
	if(num_common_species > 0){
		diff1 /= num_common_species;
	}
	else{
		THROW_WARN("The models do not share any species!");
	}
 
	delete results1;
	delete results2;
	delete integrator;
	
	vector<double> result;
	result.push_back(diff1);
	result.push_back(diff2);
	return result;
}
