/************************************************************************
  			centralcontrol.h - SBMLEvolver - CSB, FSU Jena

This software is free for use and modification, but comes without any
warranty. Usage is entirely at your own risk!

Developers: 	Thorsten Lenser (thlenser@inf.uni-jena.de)

   $Id: centralcontrol.h 3 2007-06-04 12:59:46Z thlenser $ 

This class provides the central user interface and "main" function.
**************************************************************************/
 

#ifndef CENTRALCONTROL_H
#define CENTRALCONTROL_H

#include <stdio.h>
#include <stdexcept>
#include <fstream>
#include <set>
#include <string.h>
#include <sstream>
#include <csignal>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "include.h"
#include "paramobj.h"
#include "population.h"
#include "individual.h"
#include "buildingblocks.h"
#include "mutator.h"
#include "integrator.h"

#ifdef USE_MPI_VERSION
#include <mpi.h>
#include <cmath>
#endif


using namespace std;

namespace sbmlevolver{

/**
 * \brief Central UI
 *
 * This class provides the central user interface and "main" function.
 */
class CentralControl {
public:
	/// Default constructor
	CentralControl();
	
	/// Constructor reading command line arguments
	CentralControl(int argc, char *argv[]);
	
	/// sbmlevolvers 'main' function
	void Run();
	
	/// Function to gracefully stop the run when signal SIGUSR1 is caught
	static void catchStopSignal(int sig);
	static bool stopRun; // used to stop the run when signal SIGUSR1 is received
	
protected:
	/// Parameter object contain all parameter infos for the run
	ParamObj* params;

	/// Major random number generator
	RandomNumberGenerator* rng;

#ifdef USE_MPI_VERSION
	/// In MPI mode, run as master control node
	void RunMaster();

	/// Run a population
	void RunPopulation();
	
	/// In MPI mode, run as worker-client
	void RunSlave();
#else
	/// Run (the only) population
	void RunOnSingleProcessor();
#endif
	
	/// Print the output to a string
	string generateOutput(Population*);
	
	/// Check if we have to terminate
	bool checkTermination();

	/// Helper to determine the fitness of one model
	double calculateOnlyFitness(int argc, char *argv[]);

	/// Helper to fit the parameters of a model
	void fitOnlyParameters(int argc, char *argv[]);

	/// application to study the effect of mutations
	void studyMutations(int argc, char* argv[]);
	vector<double> behaviouralDifference(Model*, Model*);


	#ifdef USE_MPI_VERSION
	/// Helper to send a string via MPI
	void sendString(int dest, string message);
	/// Helper to nonblockingly send a string via MPI
	void sendStringNonblocking(int dest, string message, MPI::Request* req1, MPI::Request* req2);
	
	/// Helper function to receive a string via MPI (from given source)
	string recvString(int* source);
	#endif
	
	int save_best_every;

	double mean_best_fit_over_all_generations;
	
	unsigned int currentTurn;
	time_t startTime;
	double bestFitness;
};


} // namespace sbmlevolver
#endif //CENTRALCONTROL_H


