/* Version information
 *
 * $Id: individual.cpp 28 2007-06-14 07:55:32Z thlenser $
 *
 */


#include "individual.h"

using namespace std;
using namespace sbmlevolver;

Individual::Individual(IndId Id, IndGen Generation, ObjectiveFunction* of, BuildingBlocks* bb) {

	// copy params
	this->Id = Id;
	this->Generation = Generation;
	
	// fitness
	this->isFit = false;
	this->Fitness = 0.0;
	//this->numFitnessEvals = 0;
	
	this->model = NULL;
	
	createMinimalModel(of,bb);
}

Individual::Individual(Model *m, IndId Id, IndGen Generation) {

	// copy params
	this->model = m;
	this->Id = Id;
	this->Generation = Generation;
	
	// fitness
	this->isFit = false;
	this->Fitness = 0.0;
	//this->numFitnessEvals = 0;
}

Individual::Individual(const char* filename, IndId Id, IndGen Generation, ObjectiveFunction* of) {

	// copy params
	this->Id = Id;
	this->Generation = Generation;
	this->model = NULL;
	
	// fitness
	this->isFit = false;
	this->Fitness = 0.0;
	//this->numFitnessEvals = 0;
	
	if(model) delete model;
	model = LoadSBMLModel(filename);
	
	prepareModel(of);
}

Individual::Individual(Individual *Ind) {
	// "smart" copy all
		
	if (Ind->model) {
		model = Ind->getModelCopy();
	}else{
		model = NULL;
	}
	
	isFit = Ind->isFit;
	Fitness = Ind->Fitness;
	
	Id = Ind->Id;
	Generation = Ind->Generation;	
}

Individual::~Individual() {
	// If model has a fitness, save it
	if(isFit){
		ofstream fitnesslog;
		fitnesslog.open("fitnesses.log",ios::app);
		fitnesslog << getId() << " " << getFitness() << endl;;
		fitnesslog.close();
	}

	// delete the related model object & sbml representation
	if (model) delete model;
	clearTags();
}

/*** (de-) serialization ***/

// special constructor to create ind from serialized string
Individual::Individual(string buf) {

	model = NULL;
	
	stringstream msg;
	msg.str(buf);
	
	msg >> Id;
	msg >> Generation;
	msg >> Fitness;
	msg >> isFit;
	while(msg.peek() == ' ') msg.ignore();
	string serializedModel;
	getline(msg, serializedModel, '\0');
	model = DeserializeSBMLModel(serializedModel);
	
	msg.clear();
}

string Individual::Serialize() {
	stringstream msg;
        msg << Id << " " << Generation << " " << Fitness << " " << isFit << " ";
	
	if (!model) THROW_ERROR("no model in individual!");
		
	string s = SerializeSBMLModel(model);
	msg << s;

	return msg.str();
}



/* The structural difference is defined as the
   number of species and reactions unique to one model
   */
double Individual::structuralDifference(Individual* ind){
	double diff = 0;
	bool found = false;
	for(int i=0; i < getModelPtr()->getNumSpecies(); i++){
		for(int j=0; j < ind->getModelPtr()->getNumSpecies(); j++){	
			if(getModelPtr()->getSpecies(i)->getId() != ind->getModelPtr()->getSpecies(j)->getId()){
				found = true;
				break;
			}
		}
		if(!found) diff += 1;
		found = false;
	}
	for(int i=0; i < ind->getModelPtr()->getNumSpecies(); i++){
		for(int j=0; j < getModelPtr()->getNumSpecies(); j++){	
			if(ind->getModelPtr()->getSpecies(i)->getId() == getModelPtr()->getSpecies(j)->getId()){
				found = true;
				break;
			}
		}
		if(!found) diff += 1;
		found = false;
	}
	
	for(int i=0; i < getModelPtr()->getNumReactions(); i++){
		for(int j=0; j < ind->getModelPtr()->getNumReactions(); j++){	
			if(equal(getModelPtr()->getReaction(i), ind->getModelPtr()->getReaction(j))) {
				found = true;
				break;
			}
		}
		if(!found) diff += 1;
		found = false;
	}
	for(int i=0; i < ind->getModelPtr()->getNumReactions(); i++){
		for(int j=0; j < getModelPtr()->getNumReactions(); j++){	
			if(equal(ind->getModelPtr()->getReaction(i), getModelPtr()->getReaction(j))) {
				found = true;
				break;
			}
		}
		if(!found) diff += 1;
		found = false;
	}
	
	return diff;
}

/* Second structural difference measure:
   Each species gets a tag depending on the type of 
   reactions it is connected to (method tagSpecies()).
   Then, tags are matched and the number of tags unique
   any one model is the structural difference.
   
   Remember to delete the tags afterwards using clearTags()
   */
double Individual::structuralDifference2(BuildingBlocks* bb, Individual* ind){
	if(speciesTags.empty()) tagSpecies(bb);
	if(ind->speciesTags.empty()) ind->tagSpecies(bb);
	
	set<int> candidates1;
	set<int> candidates2;
	for(int i=0; i < getModelPtr()->getNumSpecies(); i++) candidates1.insert(i);
	for(int i=0; i < ind->getModelPtr()->getNumSpecies(); i++) candidates2.insert(i);
	set<int>::iterator it = candidates1.begin();
	set<int>::iterator oldit;
	set<int>::iterator it2;
    int deleteit;
	while(it != candidates1.end()) {
		deleteit=0;
		it2 = candidates2.begin();
		while(it2 != candidates2.end()) {
			if(      speciesTags[     getModelPtr()->getSpecies(*it )->getId()] == ind->speciesTags[ind->getModelPtr()->getSpecies(*it2)->getId()]){
				candidates2.erase(it2);
                deleteit=1;
				break;
			}
           it2++;
		}
		if(deleteit){
		   oldit=it;
		   it++;
		   candidates1.erase(oldit);
		}
		else it++;
	}
	return (candidates1.size() + candidates2.size());
}

void Individual::tagSpecies(BuildingBlocks* bb){
	Model* m = getModelPtr();
	if(! speciesTags.empty()){
		THROW_ERROR("Tagging species when speciesTags is not empty, this should not happen");
	}
	
	for(int i=0; i < m->getNumSpecies(); i++){
		speciesTags.insert(make_pair(m->getSpecies(i)->getId(), vector<int>(bb->getNumReactions()+1,0)));
	}
	
	for(int i=0; i < m->getNumReactions(); i++){
		Reaction* re = m->getReaction(i);
		int id = bb->getReactionType(re) +1; // now between 0 and num_of_possible_reactions
		
		for(int j=0; j < re->getNumReactants(); j++){
			speciesTags[re->getReactant(j)->getSpecies()].at(id) += 1;
		}
		for(int j=0; j < re->getNumProducts(); j++){
			speciesTags[re->getProduct(j)->getSpecies()].at(id) += 1;
		}
		for(int j=0; j < re->getNumModifiers(); j++){
			speciesTags[re->getModifier(j)->getSpecies()].at(id) += 1;
		}
	}
}

void Individual::clearTags(){
	map<string, vector <int> >::iterator it;
	for(it = speciesTags.begin(); it != speciesTags.end(); it++){
		it->second.clear();
	}
	speciesTags.clear();
}

    
/*** model access ***/

Model *Individual::getModelCopy() {
	return (Model*) (getModelPtr())->clone();
}

Model *Individual::getModelPtr() {
	if (!model) {
	  THROW_ERROR("No model is available");
	}
	return model;
}
	

    
/*** fitness related ***/

bool Individual::isFitness() {
	return isFit;
}


/*
 * Calculate a new fitness for the ind
 */
void Individual::calculateFitness(ParameterFitter* paraFit, ObjectiveFunction* F) {
	// how good can we fit the model? 
	paraFit->fitParameters(getModelPtr(), F, true);
	Fitness = F->evaluate(getModelPtr());
	isFit = true;
}


//Set an externally calculated fitness
void Individual::setFitness(double fit){
	isFit = true;
	Fitness = fit;
}

void Individual::unsetFitness() {
	isFit = false;
	Fitness = 0.0;
}

//get the fitness of the individual
double Individual::getFitness() {
        if(!isFit) THROW_ERROR("Fitness has not been calculated for this model!");
		
	return Fitness;
}

    
/*** evolutionary statistics ***/

IndId Individual::getId() {
	return Id;
}

IndGen Individual::getGeneration() {
	return Generation;
}

void Individual::setId(int id){
	Id = id;
}

void Individual::setGeneration(int gen){
	Generation = gen;
}


/*** serialization ***/

void Individual::WriteToFile(char *filename /* = NULL */) {
	
	if (!filename) {
		// create own filename
		
		filename = (char*) malloc(255);
		
		if (isFitness()) {
			snprintf(filename, 255, "result_%d_%lf.model.xml", getId(), getFitness());
		}else{
			snprintf(filename, 255, "result_%d_NoFit.model.xml", getId());
		}
	}
	
	saveModel(filename);
	
	free(filename);
}


void Individual::saveModel(char *file_name) {
	SaveSBMLModel(getModelPtr(), file_name);
}


/*
 * Give the prefix DO_NOT_MUTATE to all input and output species attributes (NOT ids!!).
 * Also set the boundaryCondition attribute of the input species to "true" (disabled at the moment) and
 * set the output species' concentrations to 0.
 */
void Individual::prepareModel(ObjectiveFunction* of){
	Model* m = getModelPtr();
	
	// Check if initialAmount is set for any species. If so, set to initialConcentration
	Species* sp;
	/*	for(int i=0; i < m->getNumSpecies(); i++){
		sp = m->getSpecies(i);
		if(sp->isSetInitialAmount()){
			sp->setInitialConcentration(sp->getInitialAmount());
			THROW_WARN("The initialAmount of a species was set. This was converted to initialConcentration!");
		}
		}*/
	
	// Input species
	sp = NULL;
	for(int i=0; i < of->getNumOfInputSpecies(); i++){
		string id = of->getNameOfInputSpecies(i);
		if((id.size() > 9) && (id.substr(0,9) == string("CONSTANT_"))){
			Parameter* par = m->getParameter(id.substr(9));
			if(par == NULL){
				cerr << id << endl;
				THROW_ERROR("Input parameter not found");
			}
			else{
				setMutational(par,false);
			}
		}
		else{
			sp = m->getSpecies(id);
			if(sp == NULL){ THROW_ERROR("Input species not found!");}
			else{
				setMutational(sp,false);
				//sp->setBoundaryCondition(true);
			}
		}
	}
	
	// Output species
	for(int i=0; i < of->getNumOfOutputSpecies(); i++){
		if(of->getNameOfOutputSpecies(i) != string("ANY")){
			sp = m->getSpecies(of->getNameOfOutputSpecies(i));
			if(sp == NULL){ THROW_ERROR("Output species not found!");}
			else{
				setMutational(sp,false);
				//sp->setInitialConcentration(0);
			}
		}
	}
	sp = NULL;
	

	// Set boundaries to all unbounded parameters and initial concentrations, and initialise the stepsizes
	Parameter* par;
	Reaction* re;
	for(int i=0; i < m->getNumParameters(); i++){
		par = m->getParameter(i);
		if(!isSetAttr(par,"lower_bound")) setAttr(par,"lower_bound",0);
		if(!isSetAttr(par,"upper_bound")) setAttr(par,"upper_bound",LARGEST_PARAMETER_STRING);
	}
	for(int i=0; i < m->getNumReactions(); i++){
		re = m->getReaction(i);
		for(int j=0; j < re->getKineticLaw()->getNumParameters(); j++){
			par = re->getKineticLaw()->getParameter(j);
			if(!isSetAttr(par,"lower_bound")) setAttr(par,"lower_bound",0);
			if(!isSetAttr(par,"upper_bound")) setAttr(par,"upper_bound",LARGEST_PARAMETER_STRING);
		}
	}
	par = NULL;
	for(int i=0; i < m->getNumSpecies(); i++){
		sp = m->getSpecies(i);
		if(!isSetAttr(sp,"lower_bound")) setAttr(sp,"lower_bound",0);
		if(!isSetAttr(sp,string("upper_bound"))){
			setAttr(sp,"upper_bound",LARGEST_INITIAL_CONCENTRATION_STRING);
		}
	}
}


/* 
 * Create a minimal model with the input and output species set
 */
void Individual::createMinimalModel(ObjectiveFunction* of, BuildingBlocks* bb){
	
	// invalidate fitness
	isFit = false;
	Fitness = 0.0;
	
	if(model) delete model;
	model = new Model();
	Compartment* newcompartment = model->createCompartment();
	newcompartment->setId("compartment");
	newcompartment->setSize(1.0);
	newcompartment->setSpatialDimensions(2);
	newcompartment = NULL;
	Species* newspecies;
	for(int i=0; i < of->getNumOfInputSpecies(); i++){
		newspecies = model->createSpecies();
		newspecies->setId(of->getNameOfInputSpecies(i));
		setAttr(newspecies,"lower_bound",bb->getLowestInitialConcentration());
		setAttr(newspecies,"upper_bound",bb->getHighestInitialConcentration());
		setMutational(newspecies,false);
		newspecies->setCompartment(model->getCompartment(0)->getId());
		newspecies->setInitialConcentration(5); // Arbitrary value
		newspecies = NULL;
	}
	for(int i=0; i < of->getNumOfOutputSpecies(); i++){
		if(of->getNameOfOutputSpecies(i) != string("ANY")){
			newspecies = new Species(of->getNameOfOutputSpecies(i));	
			setMutational(newspecies,false);
		}
		else{
			newspecies = bb->getSuitableSpecies(getModelPtr());	
		}
		setAttr(newspecies,"lower_bound",bb->getLowestInitialConcentration());
		setAttr(newspecies,"upper_bound",bb->getHighestInitialConcentration());
		newspecies->setCompartment(model->getCompartment(0)->getId());
		newspecies->setInitialConcentration(0);
		model->addSpecies(newspecies);
		delete newspecies;
		newspecies = NULL;
	}

	prepareModel(of);
}


