/************************************************************************
  			individual.h - SBMLEvolver - CSB, FSU Jena

This software is free for use and modification, but comes without any
warranty. Usage is entirely at your own risk!

Developers: 	Thorsten Lenser (thlenser@inf.uni-jena.de)
		Jan Huwald (jh@sotun.de)

	Individual encapsulates the model itself as well as all parameters
	and methods related to it. Thus, it can be used to easily 
	change the representation of solution candidates.
	
	$Id: individual.h 15 2007-06-07 07:43:17Z thlenser $

**************************************************************************/

#ifndef INDIVIDUAL_H
#define INDIVIDUAL_H

#include <vector>
#include <map>
#include <set>
#include <fstream>
#include <sstream>

#include "SBMLWriter.h"
#include "SBMLReader.h"
#include "SBMLDocument.h"

#include "include.h"
#include "objectivefunction.h"
#include "buildingblocks.h"
#include "parameterfitter.h"




using namespace std;

namespace sbmlevolver{

/* Id is used in population and parallelization to address an unique individual
   HINT: as it is determined by continous incrementation it determines the max
   number of generations ! */
typedef int IndId;

/* IndGen stores a generation
   it is used for statistic issues (it determines how old a genotype is) */
typedef int IndGen;

/**
 * \brief An individual of the EA
 *
 * Individual encapsulates the model itself as well as all parameters 
 * and methods related to it.
 */
class Individual {
	public:
	/// Constructor taking only ID and generation, creating minimal model
	Individual(IndId Id, IndGen Generation, ObjectiveFunction* of, BuildingBlocks* bb);
	
	/// Constructor taking model, ID and generation
	Individual(Model *m, IndId Id, IndGen Generation);
	
	/// Constructor taking model filename, ID and generation
	Individual(const char* filename, IndId Id, IndGen Generation, ObjectiveFunction* of);
	
	/// Copy constructor
	Individual(Individual *Ind);
	
	/// Destructor
	~Individual();
	
	/// Deserialization
	Individual(string);
	
	/// Serialization
	string Serialize();
	
	/// Save the individual
	void WriteToFile(char *filename = NULL);

	/// Calculate the structural difference as the number of species and reactions unique to one model
	double structuralDifference(Individual* ind);
	double structuralDifference2(BuildingBlocks* bb, Individual* ind);
	void tagSpecies(BuildingBlocks* bb);
	void clearTags();
	
	/// get a copy of the SBML model
	Model *getModelCopy();
	
	/// get a pointer to the SBML model 
	Model *getModelPtr();	
	
	/// Test if fitness has been calculated 
	bool isFitness();	// true, if fitness is tested
	
	/// get the model fitness
	double getFitness(); 
	
	/// calculate fitness value (and store in the individual)
	void calculateFitness(ParameterFitter* paraFit, ObjectiveFunction* F);// calculate new fitness (always)
	
	/// set the fitness of the individual
	void setFitness(double);// set externally calculated fitness
	
	/// delete the individual's fitness
	void unsetFitness();
	
	// population related
	/// get ID
	IndId getId();
	
	/// Set ID
	void setId(int);
	
	/// get generation
	IndGen getGeneration();
	
	/// set generation
	void setGeneration(int);
		
protected:
	/// sbml model
	Model *model;
	
	/// state of fitness (has been tested or not)
	bool isFit;
	
	/// the models adjusted fitness
	double Fitness; 

	/// For structural difference
	map< string, vector <int> > speciesTags;
	
	// evolutionary history
	/// unique number; correlates to the absolute order of creation
	IndId Id;	
	/// generation the individual was created
    	IndGen Generation;
	

	// read and write sbml models to files
	
	/// save Model to a file
	void saveModel(char *file_name);
	
	/// Prepare the model for evolution
	void prepareModel(ObjectiveFunction*);
	
	/// Create a minimal model that contains the input and output species from the objective function
	void createMinimalModel(ObjectiveFunction*, BuildingBlocks*);
};

};

#endif
