/* Version information
 *
 * $Id: mutator.cpp 29 2007-06-15 13:35:45Z thlenser $
 *
 */


#include "mutator.h"

using namespace std;
using namespace sbmlevolver;

/**
 * Constructor
 * mutProbs is given as vector of probabilities, not distribution function!
 */
Mutator::Mutator(long int seed, BuildingBlocks* bb_ptr, vector<double> mutProbs, int maxSimMut):
	rng(seed)
{
	buildingBlocks = bb_ptr;
	
	maxSimultanMutations = maxSimMut;
	
	// Set the probabilities of the different mutations
	// (These should probably themselves be subject to the EA...)
	// For convinience, these are stored as a distribution function
	double mutSoFar = 0;
	if(mutProbs.empty()){
		for(int i=1; i <= 6; i++){
			mutSoFar += 1.0;
			mutationProbs.push_back(mutSoFar);
		}
		mutSoFar += 6.0;
		mutationProbs.push_back(6.0);
	}
	else if(mutProbs.size() != 7){
		THROW_ERROR("Wrong number of mutation probabilities given!");
	}
	else{
		for(int i=0; i < mutProbs.size(); i++){
			mutSoFar += mutProbs.at(i);
			mutationProbs.push_back(mutSoFar);
		}
	}
	
	// Normalise mutation probabilities
	for(int i=0; i < mutationProbs.size(); i++){
		mutationProbs.at(i) /= mutSoFar;
	}
}




/******************************************************************
 * In a first layer, the wrappers that work on individuals are given here.
 ****************************************************************/

/*** Recombination with another individual ***/
Individual* Mutator::recombine(Individual* ind1, Individual* ind2, IndId newId, IndGen newGeneration){
	Individual* child;
	
	child = new Individual(combineModels(ind1->getModelPtr(), ind2->getModelPtr()), newId, newGeneration);	

	ofstream hist;
	hist.open("history.log", ios::app);
	hist << ind1->getId() << " " << ind2->getId() << " " << newId;
	hist << " " << "recomb" << endl;
	hist.close();

	return child;
}


Individual* Mutator::getMutant(Individual* ind, IndId newId, IndGen newGeneration) {
	Individual* child = new Individual(ind);
	child->setId(newId);
	child->setGeneration(newGeneration);
	child->unsetFitness();
	vector<int> mutations;
	mutations = mutateModel(child->getModelPtr());
	ofstream hist;
	hist.open("history.log", ios::app);
	hist << ind->getId() << " " << newId;
	for(int i=0; i < mutations.size(); i++) hist << " " << mutations.at(i);
	hist << endl;
	hist.close();
	return child;
}

    
void Mutator::createRandomModel(Individual* ind, int num_species, int num_reactions){
	
	if(num_species == -1) num_species = buildingBlocks->getMinNumSpecies();
	if(num_reactions == -1) num_reactions = buildingBlocks->getMinNumReactions();
	
	Model* model = ind->getModelPtr();
	for(int i=(int)model->getNumSpecies(); i < num_species; i++){
		addSpeciesToModel(model); // don't use while here!
	}
	for(int i=(int)model->getNumReactions(); i < num_reactions; i++){
		addReactionToModel(model);
	}
}




/**************************************************************
 * In a second layer, the real work is done on models, below.
 **************************************************************/


/**
 * According the internal set of probabilities, maximally 
 * numMutations mutations are carried out on the model.
 */
vector<int> Mutator::mutateModel (Model* m) {
	int numMutations = maxSimultanMutations;
	if(numMutations < 1) THROW_ERROR("no mutations requested, impossible!");
	int numMutationsHere = rng.uniform(1,numMutations);
	
	vector<int> mutLog;
	mutLog.reserve(numMutationsHere);
	for(int mutNum=0; mutNum < numMutationsHere; mutNum++){
		double rand_dbl = rng.uniform();

		if(rand_dbl < mutationProbs[0]){
			// do nothing
			//cout << "Just copying model" << endl;
			mutLog.push_back(0);
		}
		else if(rand_dbl < mutationProbs[1]){
			// Add a new species
			//cout << "Trying to add species" << endl;
			if(addSpeciesToModel(m)) mutLog.push_back(1);
			else mutLog.push_back(-1);
		}
		else if(rand_dbl < mutationProbs[2]){
			//cout << "Trying to delete species" << endl;
			// Delete a species
			if(deleteSpeciesFromModel(m)) mutLog.push_back(2);
			else mutLog.push_back(-2);
		}
		else if(rand_dbl < mutationProbs[3]){
			// Add a reaction
			if(addReactionToModel(m)){
				//cout << "Added reaction" << endl;
				mutLog.push_back(3);
			}
			else{
				//cout << "Could not add reaction" << endl;
				mutLog.push_back(-3);
			}
				
		}
		else if(rand_dbl < mutationProbs[4]){
			// Delete a reaction
			if(deleteReactionFromModel(m)){
				//cout << "Deleted reaction" << endl;
				mutLog.push_back(4);
			}
			else{
				//cout << "Could not delete reaction reaction" << endl;
				mutLog.push_back(-4);
			}
		}
		else if(rand_dbl < mutationProbs[5]){
			// Replace a reaction
			//cout << "Replacing reaction" << endl;
			replaceReaction(m);
			mutLog.push_back(5);
		}
		else{ //if(rand_dbl < mutationProbs[6]){
			// Species duplication
			//cout << "Duplicating species" << endl;
			if(speciesDuplication(m)) mutLog.push_back(6);
			else mutLog.push_back(-6);
		}
	        /*else{
			// Mutate a parameter
			if(mutateParameter(m)) mutLog.push_back(7);
			else mutLog.push_back(-7);
			}*/
	}
	
	//cleanStoichiometries(m);
	
	return mutLog;
}


	
Model* Mutator::combineModels(Model* m1, Model* m2){
	// Copy model m1 first
	Model* m = (Model*) m1->clone();

	// Remove reactions
	Reaction* re;
	while(m->getNumReactions() > 0){
		re = static_cast<Reaction*> (m->getListOfReactions()->remove(0));
		delete re;
	}

	// Add all additional species from m2 to m
	for(int i=0; i < m2->getNumSpecies(); i++){
		if(m->getSpecies(m2->getSpecies(i)->getId()) == NULL){
			m->addSpecies(m2->getSpecies(i));
		}
	}

	// As a simple crossover, copy each reaction from m1 and m2 with prob 0.5
	// In the first loop, take all non-mutational reactions
	for(int i=0; i < m2->getNumReactions(); i++){
		if( (!isMutational(m2->getReaction(i))) || (rng.challenge(0.5) ) ) {
			re = (Reaction*) (m2->getReaction(i))->clone();
			re->setId(buildingBlocks->getSuitableReactionId(m));
			if(! re->getId().empty()) m->addReaction(re);
			else delete re;
			delete re;
		}
	}
	// In the second loop, ignore all mutational reactions
	for(int i=0; i < m1->getNumReactions(); i++){
		if(isMutational(m1->getReaction(i))){
			if(rng.challenge(0.5)){
				re = (Reaction*) (m1->getReaction(i))->clone();
				re->setId(buildingBlocks->getSuitableReactionId(m));
				if(! re->getId().empty()) m->addReaction(re);
				else delete re;
				delete re;
			}
		}
	}
		
	return m;
}

	

	
/**
 * Choose a parameter at random (uniformly over all parameters) and
 * add a gaussian random variable with variance 0.1
 * Also mutate initial Concentrations
 */
// DISABLED for the moment to ensure greater comparability for experiments
/*bool Mutator::mutateParameter(Model* m){
	int numParameters = 0;
	int parameterCounter = 0;

	// find the total number of parameters
	numParameters += m->getNumParameters();
	for(int i=0; i < m->getNumReactions(); i++){
		numParameters += m->getReaction(i)->getKineticLaw()->getNumParameters();
	}
	for(int i=0; i < m->getNumSpecies(); i++){
		if(isMutational(m->getSpecies(i))){
			numParameters += 1;
		}
	}
	
	// Now mutate one of them
	double rand_dbl = rng.uniform();
	Parameter* par;
	for(int i=0; i < m->getNumParameters(); i++){
		parameterCounter++;
		if(rand_dbl < ((double)parameterCounter/numParameters)){
			par = m->getParameter(i);
			par->setValue(fabs(rng.gaussian(m->getParameter(i)->getValue(), 0.1)));
			
			// Respect parameter boundaries			
			if(isSetAttr(par,"lower_bound") && (getAttr(par,"lower_bound") > par->getValue()))
				par->setValue(getAttr(par,"lower_bound"));
			if(isSetAttr(par,"upper_bound") && (getAttr(par,"upper_bound") < par->getValue()))
				par->setValue(getAttr(par,"upper_bound"));
			return true;
		}
	}
	for(int i=0; i < m->getNumReactions(); i++){
		for(int j=0; j < m->getReaction(i)->getKineticLaw()->getNumParameters(); j++){
			parameterCounter++;
			if(rand_dbl < ((double)parameterCounter/numParameters)){
				par = m->getReaction(i)->getKineticLaw()->getParameter(j); 
				par->setValue(fabs(rng.gaussian(par->getValue(), 0.1)));
				// Respect parameter boundaries
				if(isSetAttr(par,"lower_bound") && (getAttr(par,"lower_bound") > par->getValue()))
					par->setValue(getAttr(par,"lower_bound"));
				if(isSetAttr(par,"upper_bound") && (getAttr(par,"upper_bound") < par->getValue()))
					par->setValue(getAttr(par,"upper_bound"));
				return true;
			}
		}
	}
	
	Species* sp;
	for(int i=0; i < m->getNumSpecies(); i++){
		parameterCounter++;
		if(rand_dbl < ((double)parameterCounter/numParameters)){
			sp = m->getSpecies(i);
			sp->setInitialConcentration(fabs(rng.gaussian(m->getSpecies(i)->getInitialConcentration(), 0.1)));
			// Respect parameter boundaries
			if(isSetAttr(sp,"lower_bound") && (getAttr(sp,"lower_bound") > sp->getInitialConcentration()))
				sp->setInitialConcentration(getAttr(sp,"lower_bound"));
			if(isSetAttr(sp,"upper_bound") && (getAttr(sp,"upper_bound") < sp->getInitialConcentration()))
				sp->setInitialConcentration(getAttr(sp,"upper_bound"));
			return true;
		}
	}
	
	
	return false; // Nothing happend so far, no parameter mutated
	}*/



/**
 * Implementation of a "gene duplication" by
 * - choose a protein to be duplicated
 * - add a new protein
 * - copy all reactions of the old protein, replacing the old protein with the new one
 * - set all reaction rates creating old and new protein to 1/2 their value
 * 
 * return true if successful, false otherwise
 */
bool Mutator::speciesDuplication(Model* m, int species_num){
	if(m->getNumSpecies() == 0) return false;
	
	Species* old_species;
	if(species_num == -1){
		old_species = m->getSpecies(rng.uniform(0,(m->getNumSpecies() -1)));
	}
	else{
		if((species_num >= 0) && (species_num < m->getNumSpecies())){
			old_species = m->getSpecies(species_num);
		}
		else{
			THROW_ERROR("Wrong species number!");
		}
	}
	string oldSpeciesId = old_species->getId();
	
	// Check if a new species id is available in the building blocks
	if(m->getNumSpecies() >= buildingBlocks->getMaxNumSpecies()){
		#ifdef SBML_DEBUG_MUTATOR
		THROW_WARN("Cannot add species because maximum number of species in buildingBlocks is reached!\n");
		#endif
		return false;
	}
	
	Species* s = buildingBlocks->getSuitableSpecies(m);
	if(s == NULL){
		#ifdef SBML_DEBUG_MUTATOR
		THROW_WARN("Cannot add species because no suitable additional species can be found in buildingBlocks!\n");
		#endif
		return false;
	}
	string speciesId = s->getId();
	
	s->setCompartment(old_species->getCompartment());
	
	old_species->setInitialConcentration(old_species->getInitialConcentration() / 2.0);
	s->setInitialConcentration(old_species->getInitialConcentration());
	//s->setBoundaryCondition(old_species->getBoundaryCondition());
	s->setBoundaryCondition(false); // new species cannot be boundary species
	if(old_species->isSetName()) s->setName(old_species->getName());
	
	m->addSpecies(s);
	delete s;
	s = NULL;
	
	Reaction* old_re;
	for(int i=0; i < m->getNumReactions(); i++){
		old_re = m->getReaction(i);
		
		if(old_re->getReactant(oldSpeciesId) != NULL){
			Reaction* re = (Reaction*) old_re->clone();
			re->setId(buildingBlocks->getSuitableReactionId(m));
			if(re->getId().empty()) break;
			
			replaceSpeciesInReaction(re, oldSpeciesId, speciesId);
			m->addReaction(re);
			delete re;
		}
		else if(old_re->getProduct(oldSpeciesId) != NULL){
			Reaction* re = (Reaction*) old_re->clone();
			re->setId(buildingBlocks->getSuitableReactionId(m));
			if(re->getId().empty()) break;
			
			replaceSpeciesInReaction(re, oldSpeciesId, speciesId);
						
			// Half the parameter values of both reactions
			if(old_re->getKineticLaw()->getNumParameters() > 0){
				Parameter* par = old_re->getKineticLaw()->getParameter(0); // Assume first one determines reaction speed
				if(par->isSetValue()) par->setValue(0.5 * par->getValue());
				par = re->getKineticLaw()->getParameter(0); // Assume it's the first one again
				if(par->isSetValue()) par->setValue(0.5 * par->getValue());
			}
			
			m->addReaction(re);
			delete re;
		}
		else if(old_re->getModifier(oldSpeciesId) != NULL){
			Reaction* re = (Reaction*) old_re->clone();
			re->setId(buildingBlocks->getSuitableReactionId(m));
			if((re->getId()).empty()) break;
			
			replaceSpeciesInReaction(re, oldSpeciesId, speciesId);
			m->addReaction(re);
			delete re;
		}	
	}	
	return true;
}




/**
 * Add a new species from the list of globally available species to the model.
 */
bool Mutator::addSpeciesToModel (Model* m) {
	
	Species* s = buildingBlocks->getSuitableSpecies(m);
	if(s == NULL){
		return false;
	}

	int randCompartment = rng.uniform(0, m->getNumCompartments()-1);
	s->setCompartment(m->getCompartment(randCompartment)->getId());
	
	m->addSpecies(s);
	delete s;

	return true;
}


/**
 * Delete a randomly chosen species from the model.
 * All reactions are searched for participations of that species, 
 * and the species is replace by another on in these reactions. 
 */
bool Mutator::deleteSpeciesFromModel (Model* m, int species_num) {
	// Check if there is any species left in the model
	if(m->getNumSpecies() < 1) return false;

	// Check if we can delete a species
	if(m->getNumSpecies() < (buildingBlocks->getMinNumSpecies() +1)) return false;

		
	bool foundSpecies = false;
	int rand_int;
	int numSpecies = m->getNumSpecies();
	if(species_num == -1){
		for(int i=0; i < numSpecies; i++){
			// Choose the species
			rand_int = rng.uniform(0, numSpecies -1);
		
			// Check if the species can be deleted
			if(isMutational(m->getSpecies(rand_int))){
				foundSpecies = true;
				break;
			}
		}
	
		if(! foundSpecies){
			return false;
		}
	}
	else{
		if((species_num >= 0) && (species_num < numSpecies)){
			rand_int = species_num;
			if(!isMutational(m->getSpecies(rand_int))) return false;
		}
		else THROW_ERROR("Wrong species number!");
	}
	
	// Find reactions where it participates and replace it there
	string doomed_species = m->getSpecies(rand_int)->getId();

	// store the compartment of the doomed species to make sure you only replace with correct species
	string compartment = m->getSpecies(rand_int)->getCompartment();

	// delete the species from the model
	Species* sp = static_cast<Species*> (m->getListOfSpecies()->remove(rand_int));
	if(sp == NULL){
		THROW_ERROR("Warning: Trying to delete a species that is not there: in Mutator::deleteSpeciesFromModel");
	}
	delete sp;

	int randint;

	// Replace species in reactions
	string replacement_species;
	Reaction* re;
	for(int i=0; i < m->getNumReactions(); i++){
		if(m->getNumSpecies() > 0){
			re = m->getReaction(i);
			bool transport_reaction = isSetAttr(re,"transport_reaction");
			for(int j=0; j < re->getNumReactants(); j++){
				if(re->getReactant(j)->getSpecies() == doomed_species){
					if(transport_reaction){
						randint = rng.uniform(0, m->getNumSpecies()-1);
					}
					else{
						vector<int> speciesInCompartment = getSpeciesInCompartment(m,compartment);
						if(speciesInCompartment.size() == 0){
							deleteReactionFromModel(m,i);
							i--;
							continue;
						}
						else{
							randint = speciesInCompartment.at(rng.uniform(0, speciesInCompartment.size() -1));
						}
					}
					replacement_species = m->getSpecies(randint)->getId();
					replaceSpeciesInReaction(re, doomed_species, replacement_species);
				}
			}
			for(int j=0; j < re->getNumProducts(); j++){
				if(re->getProduct(j)->getSpecies() == doomed_species){
					if(transport_reaction){
						randint = rng.uniform(0, m->getNumSpecies()-1);
					}
					else{
						vector<int> speciesInCompartment = getSpeciesInCompartment(m,compartment);
						if(speciesInCompartment.size() == 0){
							deleteReactionFromModel(m,i);
							i--;
							continue;
						}
						else{
							randint = speciesInCompartment.at(rng.uniform(0, speciesInCompartment.size() -1));
						}
					}
					replacement_species = m->getSpecies(randint)->getId();
					replaceSpeciesInReaction(re, doomed_species, replacement_species);
				}
			}
			for(int j=0; j < re->getNumModifiers(); j++){
				if(re->getModifier(j)->getSpecies() == doomed_species){
					if(transport_reaction){
						randint = rng.uniform(0, m->getNumSpecies()-1);
					}
					else{
						vector<int> speciesInCompartment = getSpeciesInCompartment(m,compartment);
						if(speciesInCompartment.size() == 0){
							deleteReactionFromModel(m,i);
							i--;
							continue;
						}
						else{
							randint = speciesInCompartment.at(rng.uniform(0, speciesInCompartment.size() -1));
						}
					}
					replacement_species = m->getSpecies(randint)->getId();
					replaceSpeciesInReaction(re, doomed_species, replacement_species);
				}
			}
		}
		else{
			// Delete all reactions
			deleteReactionFromModel(m,i);
			i--;
		}
	}
	
	return true;
}


/**
 * Choose a reaction. Replace it with a different reaction.
 * Adjust the number of participants accordingly, by removing and adding species.
 */
bool Mutator::replaceReaction(Model* m){
	if(buildingBlocks->getNumReactions() == 0) THROW_ERROR("Cannot mutate without any reactions in building blocks");
	if(m->getNumSpecies() == 0) return false; //cannot have a reaction without species
	if(m->getNumReactions() == 0) return false;  //cannot replace any reaction

	bool found = false;
	int randReaction;
	Reaction* re;
	for(int i=0; i < m->getNumReactions(); i++){
		randReaction = rng.uniform(0,m->getNumReactions()-1);
		re = m->getReaction(randReaction);

		if(isMutational(re)){
			found = true;
			break;
		}
	}

	if(!found) return false;

	// Create new reaction
	// check if the reaction goes across compartments and replace with a like-wise reaction
	Reaction* re2;
	if(isSetAttr(re,"transport_reaction")){
		re2 = (Reaction*) (buildingBlocks->getTransportReaction(rng.uniform(0, buildingBlocks->getNumTransportReactions()-1)))->clone();		
	}
	else{
		re2 = (Reaction*) (buildingBlocks->getNonTransportReaction(rng.uniform(0, buildingBlocks->getNumNonTransportReactions()-1)))->clone();		
	}

	// Now try to insert all species of re into open places in re2
	
	vector<int> perm = createPermutation(re->getNumReactants());
	int k=0;
	for(int i=0; i < re2->getNumReactants(); i++){
		if(re2->getReactant(i)->getSpecies().find("Reactant") == 0){
			if(k < perm.size()){
				replaceSpeciesInReaction(re2, re2->getReactant(i)->getSpecies(), re->getReactant(perm.at(k))->getSpecies());
			}
			else{
				Species* new_sp;
				if(!isSetAttr(re2,"transport_reaction")){
					vector<int> speciesInSameCompartment = getSpeciesInCompartment(m,getCompartmentOfReaction(m,re));
					new_sp = m->getSpecies(speciesInSameCompartment.at(rng.uniform(0,speciesInSameCompartment.size()-1)));
				}
				else{
					new_sp = m->getSpecies(rng.uniform(0,m->getNumSpecies()));
				}
				replaceSpeciesInReaction(re2, re2->getReactant(i)->getSpecies(), new_sp->getId());
			}
			k++;
		}
	}
	perm = createPermutation(re->getNumProducts());
	k=0;
	for(int i=0; i < re2->getNumProducts(); i++){
		if(re2->getProduct(i)->getSpecies().find("Product") == 0){
			if(k < perm.size()){
				replaceSpeciesInReaction(re2, re2->getProduct(i)->getSpecies(), re->getProduct(perm.at(k))->getSpecies());
			}
			else{
				Species* new_sp;
				if(!isSetAttr(re2,"transport_reaction")){
					vector<int> speciesInSameCompartment = getSpeciesInCompartment(m,getCompartmentOfReaction(m,re));
					new_sp = m->getSpecies(speciesInSameCompartment.at(rng.uniform(0,speciesInSameCompartment.size()-1)));
				}
				else{
					new_sp = m->getSpecies(rng.uniform(0,m->getNumSpecies()));
				}
				replaceSpeciesInReaction(re2, re2->getProduct(i)->getSpecies(), new_sp->getId());
			}
			k++;
		}
	}
	perm = createPermutation(re->getNumModifiers());
	k=0;
	for(int i=0; i < re2->getNumModifiers(); i++){
		if(re2->getModifier(i)->getSpecies().find("Modifier") == 0){
			if(k < perm.size()){
				replaceSpeciesInReaction(re2, re2->getModifier(i)->getSpecies(), re->getModifier(perm.at(k))->getSpecies());
			}
			else{
				Species* new_sp;
				if(!isSetAttr(re2,"transport_reaction")){
					vector<int> speciesInSameCompartment = getSpeciesInCompartment(m,getCompartmentOfReaction(m,re));
					new_sp = m->getSpecies(speciesInSameCompartment.at(rng.uniform(0,speciesInSameCompartment.size()-1)));
				}
				else{
					new_sp = m->getSpecies(rng.uniform(0,m->getNumSpecies()));
				}
				replaceSpeciesInReaction(re2, re2->getModifier(i)->getSpecies(), new_sp->getId());
			}
			k++;
		}
	}
	if(m->getNumReactions() == buildingBlocks->getMinNumReactions()){
		// add reaction first, then delete old one
		if(!addReactionToModel(m,re2)){
			SaveSBMLModel(m,"error.xml");
			THROW_ERROR("Cannot add new reaction in replaceReaction() !");
		}
		if(!deleteReactionFromModel(m,randReaction)){
			SaveSBMLModel(m,"error.xml");
			THROW_ERROR("Cannot remove old reaction in replaceReaction() !");
		}
	}
	else{
		// delete old one first, then add new one
		if(!deleteReactionFromModel(m,randReaction)){
			SaveSBMLModel(m,"error.xml");
			THROW_ERROR("Cannot remove old reaction in replaceReaction() !");
		}
		if(!addReactionToModel(m,re2)){
			SaveSBMLModel(m,"error.xml");
			THROW_ERROR("Cannot add new reaction in replaceReaction() !");
		}
	}

	return true;
}


/**
 * Choose a reaction from the building blocks
 * See if corresponding species are available
 * If so, add it
 */
bool Mutator::addReactionToModel (Model* m, Reaction* re) {	
	if(buildingBlocks->getNumReactions() == 0){
		// return false; // No reaction to add
		THROW_ERROR("Trying to add reaction when building blocks contain none");
	}
	if(m->getNumSpecies() == 0) return false; //cannot have a reaction without species

	if(re == NULL){
		re = (Reaction*) (buildingBlocks->getReaction(rng.uniform(0, buildingBlocks->getNumReactions()-1)))->clone();
	}
	
	bool transport_reaction = isSetAttr(re,"transport_reaction");

	int numReactants = re->getNumReactants();
	int numProducts = re->getNumProducts();
	int numModifiers = re->getNumModifiers();
	int randint;

	// Replace placeholders with arbitrary species
	// Also check if the other species are there, otherwise: abort
	for(int i=0; i < numReactants; i++){
		if(re->getReactant(i)->getSpecies().find("Reactant") == 0){
			if((i==0) || transport_reaction){
				randint = rng.uniform(0, m->getNumSpecies()-1);
			}
			else{
				vector<int> speciesInCompartment = getSpeciesInCompartment(m,getCompartmentOfReaction(m,re));
				randint = speciesInCompartment.at(rng.uniform(0, speciesInCompartment.size() -1));
			}
			replaceSpeciesInReaction(re, re->getReactant(i)->getSpecies(), m->getSpecies(randint)->getId());
		}
		else{
			if(m->getSpecies(re->getReactant(i)->getSpecies()) == NULL){
				THROW_ERROR("Trying to add a reaction with a species that does not exist in the model!");
				delete re;
				return false;
			}
		}
	}
	for(int i=0; i < numModifiers; i++){
		if(re->getModifier(i)->getSpecies().find("Modifier") == 0){
			if(((re->getNumReactants() == 0) && (i=0)) || transport_reaction){
				randint = rng.uniform(0, m->getNumSpecies()-1);
			}
			else{
				vector<int> speciesInCompartment = getSpeciesInCompartment(m,getCompartmentOfReaction(m,re));
				randint = speciesInCompartment.at(rng.uniform(0, speciesInCompartment.size() -1));
			}
			replaceSpeciesInReaction(re, re->getModifier(i)->getSpecies(), m->getSpecies(randint)->getId());
		}
		else{
			if(m->getSpecies(re->getModifier(i)->getSpecies()) == NULL){
				delete re;
				return false;
			}
		}
	}
	for(int i=0; i < numProducts; i++){
		if(re->getProduct(i)->getSpecies().find("Product") == 0){
			if(((re->getNumReactants() == 0) && (re->getNumProducts() == 0) && (i=0))){
				randint = rng.uniform(0, m->getNumSpecies()-1);					
			}
			else if(transport_reaction){
				// make sure that Product0 comes from a different compartment than reactant0
				randint = rng.uniform(0, m->getNumSpecies()-1);								      
				if((i=0) && (re->getNumReactants() > 0)){
					int counter = 0;
					while((counter < 10) && (m->getSpecies(randint)->getCompartment() == m->getSpecies(re->getReactant(0)->getSpecies())->getCompartment())){
						randint = rng.uniform(0, m->getNumSpecies()-1);					
						counter++;
					}					
				}
				else{
					randint = rng.uniform(0, m->getNumSpecies()-1);					
				}
			}
			else{
				vector<int> speciesInCompartment = getSpeciesInCompartment(m,getCompartmentOfReaction(m,re));
				randint = speciesInCompartment.at(rng.uniform(0, speciesInCompartment.size() -1));
			}
			replaceSpeciesInReaction(re, re->getProduct(i)->getSpecies(), m->getSpecies(randint)->getId());
		}
		else{
			if(m->getSpecies(re->getProduct(i)->getSpecies()) == NULL){
				delete re;
				return false;
			}
		}
	}

	// Initialise parameters randomly
	if(re->isSetKineticLaw()){
		for(int i=0; i < re->getKineticLaw()->getNumParameters(); i++){
			Parameter* par = re->getKineticLaw()->getParameter(i);
			if(! par->isSetValue()){
				par->setValue(getAttr(par,"lower_bound") + (getAttr(par,"upper_bound")-getAttr(par,"lower_bound"))*rng.uniform());
			}
		}
	}
	
	// set ID
	re->setId(buildingBlocks->getSuitableReactionId(m));
	if(re->getId().empty()){
		delete re;
		return false;
	}
	
	m->addReaction(re);
	delete re;
	re = NULL;
	return true;
}


/**
 * Choose a reaction.
 * If that reaction is mutable, delete it.
 */
bool Mutator::deleteReactionFromModel (Model* m, int reaction_num) {
	// Check if at least one reaction is left in the model
	if(m->getNumReactions() < 1) return false;

	// Check if we can delete a reaction
	if(m->getNumReactions() < (buildingBlocks->getMinNumReactions() +1)) return false;
	
	int rand_reaction;
	bool foundReaction = false;
	for(int i=0; i < m->getNumReactions(); i++){
		if(reaction_num==-1) rand_reaction = rng.uniform(0, m->getNumReactions() -1);
		else if((reaction_num >= 0) && (reaction_num < m->getNumReactions())){
			rand_reaction = reaction_num;
			i = m->getNumReactions();
			if(!isMutational(m->getReaction(rand_reaction))) return false;
		}
		else THROW_ERROR("Wrong reaction number!");
	
		// Check if the reaction can be mutated
		if(isMutational(m->getReaction(rand_reaction))){
			foundReaction = true;
			break;
		}
	}
	if(! foundReaction){
		return false;
	}
	
	// Get block id of reaction (if any)
	int block_id = -1;
	if(isSetAttr(m->getReaction(rand_reaction),"block_id")) block_id = (int) getAttr(m->getReaction(rand_reaction),"block_id");
	
	// Remove the reaction from the model
	Reaction* re = static_cast<Reaction*> (m->getListOfReactions()->remove(rand_reaction));
	delete re;
	
	// Remove any other reactions with the same block id
	for(int i=0; i < m->getNumReactions(); i++){
		if(isSetAttr(m->getReaction(i),"block_id")){
			if(getAttr(m->getReaction(i),"block_id") == block_id){
				if(!deleteReactionFromModel(m,i)) THROW_ERROR("Cannot delete reaction in reaction-block of which one reaction has already been deleted, this should not happen");
				i--;
			}
		}
	}
	
	return true;
}


 



// Little helper
int factorial (int n) {
	int f = 1;

	for (int i=1; i <= n; i++)
		f = f*i;
	return f;
}

 	
/**
 * Choose a new feasible rate law for the reaction
 */
bool Mutator::chooseNewKineticLaw(Reaction* re){
	delete re->getKineticLaw();
	int numReactants = re->getNumReactants();
 	int numProducts = re->getNumProducts();
 	int numModifiers = re->getNumModifiers();
 	// Feasible kinetic laws have the same number of reactants, products and modifiers
	int numNewKineticLaws = buildingBlocks->getNumReactions(numReactants, numProducts, numModifiers);
 	
	// Go through possible reactions in a random order. Use the first one that fits
	vector<int> possibilities;
	possibilities.reserve(numNewKineticLaws);
	for(int i=0; i < numNewKineticLaws; i++){
		possibilities.push_back(i);
	}
	int tmp, rand_pos;
	for(int i=0; i < (numNewKineticLaws -1); i++){
		tmp = possibilities.at(i);
		rand_pos = rng.uniform(i,(numNewKineticLaws-1));
		possibilities.at(i) = possibilities.at(rand_pos);
		possibilities.at(rand_pos) = tmp;
	}
	
	for(int i=0; i < numNewKineticLaws; i++){
		Reaction* tmpRe = buildingBlocks->getReaction(numReactants,numProducts,numModifiers,possibilities.at(i));
		if(tmpRe == NULL) THROW_ERROR("Error in retrieving kinetic law!");
		// Now check that reactants, modifiers and products fit
		// try every permutation
		vector<int> perm;
		perm.reserve(numReactants);
		for(int j=0; j < numReactants; j++){
			perm.push_back(j);
		}
		bool successRe = true;
		vector<int> reac_placeholders;
		for(int j=0; j < factorial(numReactants); j++){
			for(int k=0; k < perm.size(); k++){
				if(re->getReactant(k)->getSpecies() != tmpRe->getReactant(perm.at(k))->getSpecies()){
					if(tmpRe->getReactant(perm.at(k))->getSpecies().find("Reactant") != 0) successRe = false;
					else reac_placeholders.push_back(k);
				}
			}
			if(successRe) break;
			successRe = true;
			reac_placeholders.clear();
			next_permutation(perm.begin(), perm.end());
		}
		
		perm.clear();
		perm.reserve(numModifiers);
		for(int j=0; j < numModifiers; j++){
			perm.push_back(j);
		}
		bool successMod = true;
		vector<int> mod_placeholders;
		for(int j=0; j < factorial(numModifiers); j++){
			for(int k=0; k < perm.size(); k++){
				if(re->getModifier(k)->getSpecies() != tmpRe->getModifier(perm.at(k))->getSpecies()){
					if(tmpRe->getModifier(perm.at(k))->getSpecies().find("Modifier") != 0) successMod = false;
					else mod_placeholders.push_back(k);
				}
			}
			if(successMod) break;
			successMod = true;
			mod_placeholders.clear();
			next_permutation(perm.begin(), perm.end());
		}
		
		perm.clear();
		perm.reserve(numProducts);
		for(int j=0; j < numProducts; j++){
			perm.push_back(j);
		}
		bool successPr = true;
		vector<int> prod_placeholders;
		for(int j=0; j < factorial(numProducts); j++){
			for(int k=0; k < perm.size(); k++){
				if(re->getProduct(k)->getSpecies() != tmpRe->getProduct(perm.at(k))->getSpecies()){
					if(tmpRe->getProduct(perm.at(k))->getSpecies().find("Product") != 0) successPr = false;	
					else prod_placeholders.push_back(k);
				}
			}
			if(successPr) break;
			successPr = true;
			prod_placeholders.clear();
			next_permutation(perm.begin(), perm.end());
		}
	
		if(successRe && successMod && successPr){
			// Now we have to copy this, so that it belongs to the real reaction
			KineticLaw* copyKl = (KineticLaw*) (tmpRe->getKineticLaw())->clone();
			re->setKineticLaw(copyKl);
			
			expandKineticLaw(re, reac_placeholders, mod_placeholders, prod_placeholders);
			re->setReversible(false);
			
			return true;
		}
	}
	
	// No kinetic law could be found, meaning such a reaction should not exist
	// Set reaction to 0, but return false, so that it can be deleted.
	KineticLaw* newKl = new KineticLaw("0");
	re->setKineticLaw(newKl);
	re->setReversible(false);
			
	return false;
}
 	
 	


void Mutator::expandKineticLaw(Reaction* re, vector<int> reac_pl, vector<int> mod_pl, vector<int> prod_pl){
	string formula = re->getKineticLaw()->getFormula();
	int index;
	
	for(int j=0; j < reac_pl.size(); j++){
		string pattern("Reactant");
		ostringstream converter;
		converter << j;
		pattern.append(converter.str());
		converter.flush();
		
		index = formula.find(pattern.c_str(),0);
		while(index != string::npos){
			formula.erase(index,pattern.size());
			formula.insert(index, re->getReactant(reac_pl.at(j))->getSpecies());
			index = formula.find(pattern,0);
		}
	}
	
	for(int j=0; j < prod_pl.size(); j++){
		string pattern("Product");
		ostringstream converter;
		converter << j;
		pattern.append(converter.str());
		converter.flush();
		
		index = formula.find(pattern.c_str(),0);
		while(index != string::npos){
			formula.erase(index,pattern.size());
			formula.insert(index, re->getProduct(prod_pl.at(j))->getSpecies());
			index = formula.find(pattern,0);
		}
	}

	for(int j=0; j < mod_pl.size(); j++){
		string pattern("Modifier");
		ostringstream converter;
		converter << j;
		pattern.append(converter.str());
		converter.flush();
		
		index = formula.find(pattern.c_str(),0);
		while(index != string::npos){
			formula.erase(index,pattern.size());
			formula.insert(index, re->getModifier(mod_pl.at(j))->getSpecies());
			index = formula.find(pattern,0);
		}
	}
	
	re->getKineticLaw()->setFormula(formula);
}


vector<int> Mutator::createPermutation(int n){
	vector<int> perm;
	perm.reserve(n);
	for(int i=0; i<n; i++){
		perm.push_back(i);
	}
	int tmp;
	int rand_pos;
	for(int i=0; i < (n-1); i++){
		tmp = perm.at(i);
		rand_pos = rng.uniform(i,(n-1));
		perm.at(i) = perm.at(rand_pos);
		perm.at(rand_pos) = tmp;
	}
	return perm;
}
