/************************************************************************
  			mutator.h - SBMLevolver - CSB, FSU Jena

This software is free for use and modification, but comes without any
warranty. Usage is entirely at your own risk!

Developers: 	Thorsten Lenser (thlenser@inf.uni-jena.de)
				Christoph Schmidt (nyota@web.de)
				Jan Huwald (jh@sotun.de)
				Lisa Hellwig (louis-chan@web.de)
				

	$Id: mutator.h 29 2007-06-15 13:35:45Z thlenser $


This class provides methods to mutate a model or recombine two models
**************************************************************************/
 

#ifndef MUTATOR_H
#define MUTATOR_H

#include <string>
#include <cmath>
#include <sstream>
#include <fstream>

#include "Species.h"
#include "Model.h"

#include "include.h"
#include "randomnumbergenerator.h"
#include "buildingblocks.h"
#include "individual.h"
#include "sbmltools.h"


using namespace std;

namespace sbmlevolver{
	

/**
 * \brief Mutates models
 *
 * This class provides methods to mutate a model
 */
class Mutator {

public:
	/// Initialise class, start RNG
	Mutator(long int, BuildingBlocks*, vector<double>, int = 1);

	/// Carry out a mutation on the model, return the type of mutation
   	vector<int> mutateModel (Model* m);

	/// recombine two individuals
	Individual* recombine(Individual* ind1, Individual* ind2, IndId newId, IndGen newGeneration);
	
	/// get a mutated copy of the individual
	Individual* getMutant(Individual* ind, IndId newId, IndGen newGeneration);

	/// Create an initial solution
	void createRandomModel(Individual* ind, int num_species = -1, int num_reactions = -1);

   	
   	/// Combine two models by copying each reaction with probability 0.5
   	Model* combineModels(Model* m1, Model* m2);
   	
   	/// Duplicate a species
	bool speciesDuplication(Model* m, int species_num = -1);

	/// Add a new species to the model
    	bool addSpeciesToModel (Model* m);
    
	/// Delete a species from the model
	bool deleteSpeciesFromModel (Model* m, int species_num = -1);
  
	/// Replace a reaction with a new reaction, maintaining most of the participants
	bool replaceReaction(Model* m);
  
    	/// Add a new reaction to the model
	bool addReactionToModel(Model* m, Reaction* re = NULL);
  
    	/// Remove a reaction from the model
    	bool deleteReactionFromModel (Model* m, int reaction_num=-1);  
	
	/// Choose a new (feasible) rate law for the reaction, return false if no kinetic law is available
	bool chooseNewKineticLaw(Reaction*);
   	
    
private:
	/// Facility to generate random numbers
	RandomNumberGenerator rng;
	
	/// Mutator has to know about all possibilities
	BuildingBlocks* buildingBlocks;
	
	/// Probabilities of the different types of mutations (probs are in order of the listing of functions)
	vector<double> mutationProbs;
	
	// Mutate any one parameter
	//bool mutateParameter(Model* m);
	    
    	/// Expand a blueprint kinetic law into a proper one with the right species names
	void expandKineticLaw(Reaction*, vector<int>, vector<int>, vector<int>);
    
	/// Maximal number of mutations done in one step (actual number is drawn uniformly between 1 and max
    	int maxSimultanMutations;
	
	/// Create a permutation of the integers 1..n
	vector<int> createPermutation(int n);
  
};


} // namespace sbmlevolver

#endif //MUTATOR_H





