/* Version information
 *
 * $Id: objectivefunction.cpp 25 2007-06-07 13:25:35Z thlenser $
 *
 */


#include "objectivefunction.h"

using namespace std;
using namespace sbmlevolver;


// Include the specific objective function
// Format is "objectives/SPECIFIC_OBJECTIVE.cpp
// The first one is the default:
#include "objectives/io-table.cpp"
//#include "objectives/flipflop.cpp"
//#include "objectives/stochastic-switching.cpp"


bool ObjectiveFunctionBase::getUseAIC(){
	return useAIC;
}
	
double ObjectiveFunctionBase::getNonMassConservancePenalty(){
	return nonMassConservancePenalty;
}


/**
 * Get the number of input species.
 * Number of different species, which are involved in the objective function
 * with their initial concentrations as input.
 */
int ObjectiveFunctionBase::getNumOfInputSpecies ( ) {
  return inputSpecies.size();
}


/**
 * Get the number of output species
 * Number of different species, which are involved in the objective function
 * with their concentrations curves as output
 */
int ObjectiveFunctionBase::getNumOfOutputSpecies ( ) {
  return outputSpecies.size();
}


/**
 * Get the name of the input species with the internal number i
 */
string ObjectiveFunctionBase::getNameOfInputSpecies(int num){
	try{
		return inputSpecies.at(num);
	}
	catch(...){
		THROW_ERROR("Input species does not exist!");
	}
}

/**
 * Get the name of the output species with the internal number i
 */
string ObjectiveFunctionBase::getNameOfOutputSpecies(int num){
	try{
		return outputSpecies.at(num);
	}
	catch(...){
		THROW_ERROR("Output species does not exist!");
	}
}

/**
 * Check if input species
 */
bool ObjectiveFunctionBase::isInputSpecies(string id){
	bool found = false;
	for(int i=0; i < getNumOfInputSpecies(); i++){
		if(id == getNameOfInputSpecies(i)) found = true;
		break;
	}
	return found;
}

/**
 * Check if output species
 */
bool ObjectiveFunctionBase::isOutputSpecies(string id){
	bool found = false;
	for(int i=0; i < getNumOfOutputSpecies(); i++){
		if(id == getNameOfOutputSpecies(i)) found = true;
		break;
	}
	return found;
}



#undef NEXT_INPUT


double ObjectiveFunctionBase::evaluate(Model* m){
	double fitness = calculateRawFitness(m);
	//cout << "Hal " << fitness << endl;
	// Take mass-conservance into account
	if(fitness < VERY_BAD_FITNESS -10){
		if(getNonMassConservancePenalty() > 0)
			fitness += (getNonMassConservancePenalty() * massConserving(m));	
		//cout << "Mass " << massConserving(m) << " Penalty " << getNonMassConservancePenalty() << endl;
		if(fitness > VERY_BAD_FITNESS) fitness = VERY_BAD_FITNESS; // Otherwise it might be too large
	}
	//cout << fitness << endl;
	// Use AIC to penalise number of parameters
	if(!getUseAIC()) ;
	else{
		// Use AIC to minimise deviation while keeping the number of parameters low
		int k = m->getNumParameters();
		k += m->getNumSpecies(); // Initial concentrations
		for(int i=0; i < m->getNumReactions(); i++){
			k += m->getReaction(i)->getKineticLaw()->getNumParameters();
		}
		int n = numOfCases * numOfTimesteps * getNumOfOutputSpecies();
		if(n <= (k+1)) fitness = VERY_BAD_FITNESS; // way too many parameters
		fitness = 2*k + (n * log(fitness)) + 2*k*(k+1)/(n-k-1);
	}
	//cout << fitness << endl;
	// Use simple size penalty to keep number of reactions and species low?
	if(sizePenalty > 0){
		fitness += sizePenalty * (m->getNumReactions() + m->getNumSpecies());
		if(fitness > VERY_BAD_FITNESS) fitness = VERY_BAD_FITNESS;
	}
//cout << fitness << " Hal" << endl;
//cout << fitness << endl;
	return fitness;
}






/* Find out if the reaction network can be mass-maintaining,
 * i.e. if there is a vector m of molecule-masses s.t. Mm=0, m>0
 * (or aquivalently m >= 1) when M is the stoichiometric matrix
 *
 * To do that, transfer the linear programming problem into 
 * restricted normal form and feed it into the simplex algorithm.
 * 
 * If the result is 0, the network is mass conserving, otherwise
 * the result is > 0
 *
 * Note: The results of this function can be quite large, so that
 * it should be capped at VERY_BAD_FITNESS if it is supposed to be
 * used as a fitness penalty.
 */
double ObjectiveFunctionBase::massConserving(Model*m ){
#ifdef DO_NOT_USE_LP_SOLVE
	THROW_ERROR("Mass conservance cannot be tested when program is compiled without lp_solve support");
#else
	int numReactions = m->getNumReactions();
	int numSpecies = m->getNumSpecies(); 
  	
  	// Set stoichiometric matrix M (elements are set to 0)
	gsl_matrix* stoichMat = gsl_matrix_calloc(numReactions, numSpecies);
	int numReactants = 0;
	int numProducts = 0;
	SpeciesReference* sr = NULL;
	Reaction* re;
	double stoich;

	for(int i=0; i < numReactions; i++){
		re = m->getReaction(i);
		numReactants = re->getNumReactants();
		numProducts = re->getNumProducts();
		for(int j=0; j < numSpecies; j++){
			sr=NULL;
  			// Find out if species is reactant
			sr = re->getReactant(m->getSpecies(j)->getId());
			if(sr){
				// deal with multiple occurrences
				stoich = 0;
				for(int k=0; k < re->getNumReactants(); k++){
					if(re->getReactant(k)->getSpecies() == m->getSpecies(j)->getId()){
						stoich += re->getReactant(k)->getStoichiometry();
					}
				}
				gsl_matrix_set(stoichMat,i,j,(- stoich));
			}
			else{
  				// Find out if species is product
				sr = re->getProduct(m->getSpecies(j)->getId());
				if(sr){
					// deal with multiple occurrences
					stoich = 0;
					for(int k=0; k < re->getNumProducts(); k++){
						if(re->getProduct(k)->getSpecies() == m->getSpecies(j)->getId()){
							stoich += re->getProduct(k)->getStoichiometry();
						}
					}
					gsl_matrix_set(stoichMat, i, j, stoich);
				}
  				// else set to 0, already done
			}
		}
	}
  	
  	// For a justification of the following procedure, see the docs
  	// Maybe use BLAS support here, to speed up computation
  	// Maybe it would be faster to solve the original matrix? Try it!
	gsl_vector* b = gsl_vector_calloc(numReactions);
	double tmp_dbl = 0;
	for(int i=0; i < numReactions; i++){
		for(int j=0; j < numSpecies; j++){
			tmp_dbl += gsl_matrix_get(stoichMat, i, j);
		}
		gsl_vector_set(b, i, tmp_dbl);
		tmp_dbl = 0;
	}
  	
  	// Now initialise the linear programming problem
	lprec *lp;
	lp = make_lp(numReactions, numReactions + numSpecies);
	set_add_rowmode(lp, FALSE);
	for(int i=1; i <= numReactions; i++){
		for(int j=1; j <= numReactions; j++){
			if(i==j) set_mat(lp,i,j,1);
			else set_mat(lp,i,j,0);
		}
		for(int j = (numReactions+1); j <= (numReactions + numSpecies); j++){
			set_mat(lp,i,j, (- gsl_matrix_get(stoichMat, i-1, (j-numReactions-1))));
		}
	}
	for(int i=1; i <= numReactions; i++){
		set_rh(lp,i,gsl_vector_get(b,i-1));
		set_constr_type(lp, i, EQ);
	}
	for(int j=1; j <= numReactions; j++){
		set_obj(lp, j, 1);
	}
	for(int j= (numReactions+1); j <= (numReactions + numSpecies); j++){
		set_obj(lp, j, 0);
	}
	set_minim(lp);
	
	//print_lp(lp);
	set_verbose(lp, CRITICAL);
	solve(lp);
	
	double lp_result = get_objective(lp);
	free_lp(&lp);
	gsl_matrix_free(stoichMat);
	gsl_vector_free(b);
	return lp_result;
#endif
}


// Determine whether a run should be stored
void ObjectiveFunctionBase::setStoreRun(bool storeRun){
	this->storeRun = storeRun;
}







/* The following is directly copied from the file sbmlResults.c of the sbml_odeSolver. Thanks! */

static timeCourseArray_t *TimeCourseArray_create(int names, int timepoints);
static timeCourse_t *TimeCourse_create(int timepoints);

/* The function
   SBMLResults SBMLResults_create(Model_t *m, int timepoints)
   allocates memory for simulation results (time courses) mapped back
   on SBML structures (i.e. species, and non-constant compartments
   and parameters. It takes CvodeData as an argument, that has to
   contain CVODE integraton results.
*/

SBMLResults_t *SBMLResults_create(Model_t *m, int timepoints)
{
	int i, num_reactions, num_species, num_compartments, num_parameters;
	Species_t *s;
	Compartment_t *c;
	Parameter_t *p;
	Reaction_t *r;
	SBMLResults_t *results;
	timeCourse_t *tc;

	ASSIGN_NEW_MEMORY(results, struct _SBMLResults, NULL);
  
	/* Allocating the time array  */
	results->time =  TimeCourse_create(timepoints);
	ASSIGN_NEW_MEMORY_BLOCK(results->time->name, 5, char, NULL);
	sprintf(results->time->name, "time");
  
	/* Allocating arrays for all SBML species */
	num_species = Model_getNumSpecies(m);
	results->species = TimeCourseArray_create(num_species, timepoints);

  
	/* Writing species names */
	for ( i=0; i<Model_getNumSpecies(m); i++) {
		s = Model_getSpecies(m, i);
		tc = results->species->tc[i];
		ASSIGN_NEW_MEMORY_BLOCK(tc->name, strlen(Species_getId(s))+1, char, NULL);
		sprintf(tc->name, "%s", Species_getId(s));
	}

	/* Allocating arrays for all variable SBML compartments */
	num_compartments = 0;
	for ( i=0; i<Model_getNumCompartments(m); i++ ) 
		if ( ! Compartment_getConstant(Model_getCompartment(m, i)) )
			num_compartments++;
  
	results->compartments = TimeCourseArray_create(num_compartments, timepoints);
	/* Writing variable compartment names */
	for ( i=0; i<Model_getNumCompartments(m); i++) {
		c = Model_getCompartment(m, i);
		if ( ! Compartment_getConstant(c) ) {
			tc = results->compartments->tc[i];
			ASSIGN_NEW_MEMORY_BLOCK(tc->name, strlen(Compartment_getId(c))+1,
					char, NULL);
			sprintf(tc->name, "%s", Compartment_getId(c));			
		}
	}

	/* Allocating arrays for all variable SBML parameters */
	num_parameters = 0;
	for ( i=0; i<Model_getNumParameters(m); i++ ) 
		if ( ! Parameter_getConstant(Model_getParameter(m, i)) )
			num_parameters++;
  
	results->parameters = TimeCourseArray_create(num_parameters, timepoints);
	/* Writing variable parameter names */
	for ( i=0; i<Model_getNumParameters(m); i++) {
		p = Model_getParameter(m, i);
		if ( ! Parameter_getConstant(p) ) {
			tc = results->parameters->tc[i];
			ASSIGN_NEW_MEMORY_BLOCK(tc->name, strlen(Parameter_getId(p))+1,
					char, NULL);
			sprintf(tc->name, "%s", Parameter_getId(p));
		}
	}  

	/* Allocating arrays for all variable SBML reactions */
	num_reactions = Model_getNumReactions(m);
	results->fluxes = TimeCourseArray_create(num_reactions, timepoints);
	/* Writing reaction names */
	for ( i=0; i<Model_getNumReactions(m); i++ ) {
		r = Model_getReaction(m, i);
		tc = results->fluxes->tc[i];
		ASSIGN_NEW_MEMORY_BLOCK(tc->name, strlen(Reaction_getId(r))+1, char, NULL);
		sprintf(tc->name, "%s", Reaction_getId(r));
	}

	return results;
}


static timeCourse_t *TimeCourse_create(int timepoints)
{
	timeCourse_t *tc;
	ASSIGN_NEW_MEMORY(tc, struct timeCourse, NULL);
	tc->timepoints = timepoints;
	/* timecourses variable matrix */
	ASSIGN_NEW_MEMORY_BLOCK(tc->values, timepoints, double, NULL);  
	return tc;
}



static timeCourseArray_t *TimeCourseArray_create(int num_val, int timepoints)
{
	int i;
	timeCourseArray_t *tcA;

	ASSIGN_NEW_MEMORY(tcA, struct timeCourseArray, NULL);
	tcA->num_val = num_val;
	/* num_val time course structures */  
	ASSIGN_NEW_MEMORY_BLOCK(tcA->tc, num_val, struct timeCourse *, NULL);
	for ( i=0; i<num_val; i++ )
		tcA->tc[i] = TimeCourse_create(timepoints);
  
	return tcA;
}
