/************************************************************************
  			objectivefunction.h - SBMLEvolver - CSB, FSU Jena

This software is free for use and modification, but comes without any
warranty. Usage is entirely at your own risk!

Developers: 	Thorsten Lenser (thlenser@inf.uni-jena.de)

This class stores the type of objective function and provides
facilities to specifiy the objective function via an
input/output table.
It also evaluates the fitness of a given model (solution candidate)

 * $Id: objectivefunction.h 3 2007-06-04 12:59:46Z thlenser $

**************************************************************************/
 

#ifndef OBJECTIVEFUNCTION_H
#define OBJECTIVEFUNCTION_H

#include <string.h>
#include <vector>

#include "SBMLTypes.h"
#include "gsl/gsl_vector.h"
#include "gsl/gsl_matrix.h"
#include "gsl/gsl_errno.h"
#include "gsl/gsl_fft_real.h"
#include "gsl/gsl_fft_halfcomplex.h"

#ifndef DO_NOT_USE_LP_SOLVE
#include "lp_lib.h"
#endif 

#include "include.h"
#include "sbmltools.h"


using namespace std;

namespace sbmlevolver{

/**
 *\brief Manages objective function and fitness evaluation
 *
 * This class provides basic tools for the objective function.
 * It is the base class of class ObjectiveFunctionBase, which is
 * implemented on a problem-specific basis on files objectives/<type>.cpp
 */
class ObjectiveFunctionBase{
public:
	/// Function that does common modifications to the fitness from problem-specific calculateRawFitness
	double evaluate(Model* m);
	
	/// Pure virtual function that is implemented in the derived class
	virtual double calculateRawFitness(Model* m) = 0;
	
	/// Check whether AIC should be used
	bool getUseAIC();
	
	/// Get the penalty for non-massconservance
	double getNonMassConservancePenalty();
	           
	/// Get the number of input species
	int getNumOfInputSpecies();       
	       
	/// Get the number of output species
	int getNumOfOutputSpecies ();  
	
	/// Return the name of input species i
	string getNameOfInputSpecies(int);
	
	/// Return the name of output species i
	string getNameOfOutputSpecies(int);
	
	/// Check if species is input species
	bool isInputSpecies(string);
	
	/// Check if species is output species
	bool isOutputSpecies(string);
	
	/// Determine whether a run should be stored
	void setStoreRun(bool storeRun);
protected:
	/*
	 * OBJECTIVE FUNCTION SPECIFICATION
 	 */

	/// Use Akaikes Information criterion (where appropriate, true=use it, default is false)
	bool useAIC;
	
	/// Penalty factor for non-massconservance (default 10000)
	double nonMassConservancePenalty;

	/// Penalty for size of model
	double sizePenalty;
	
	/// the number of fitness cases
	int numOfCases;
	
	/// the number of time steps for fitness evaluation
	int numOfTimesteps;
   
	/// the offset (ignored time) for fitness evaluation
	int offset;
  
	/// the time step for fitness evaluation
	double timestep;
   
	/// the input species
	vector<string> inputSpecies;
   
	/// the output species
	vector<string> outputSpecies;  

	/// get penalty for non-mass-conservance
	double getNonMassConservingPenalty();
	
	/// set penalty for non-mass-sonservance
	void setNonMassConservingPenalty(double);
	
	/// Does the model fulfill mass conservance? 0 = mass-conserving, > 0 degree of "non-conservance"
	double massConserving(Model*); 
	
	/// Penalty for models that do not fulfill mass-conservance
	double nonMassConservingPenalty;
	
	/// Store a run when integrating (has to be implemented by the integrator)
	bool storeRun;
};

} // namespace sbmlevolver

// Include the object declaration for the specific objective function
#include "objectives/io-table.h"
// #include "objectives/flipflop.h"
//#include "objectives/stochastic-switching.h"

#endif //OBJECTIVEFUNCTION_H




