/*
 * This class defines the specific objective function. 
 * Include this code in objectivefunction.cpp
 * Always specify the calculateRawFitness function
 */
 
//#include "../objectivefunction.h"
// Not needed, you can use anything from ObjectiveFunction

ObjectiveFunction::ObjectiveFunction(unsigned long int seed, const char* filename){
	ifstream* input;
        if(filename == NULL) input = new ifstream("objectivefunction.txt");
	else input = new ifstream(filename);
	if(!input->good()) THROW_ERROR("Could not open file objectivefunction.txt");
	
	readInput((istream*) input);
	
	delete input;
	
	// Initialise integrator
	integrator = new Integrator();
}
	

ObjectiveFunction::ObjectiveFunction(string input, unsigned long int seed){
	istringstream* inputFile = new istringstream();
	inputFile->str(input);

	readInput((istream*) inputFile);
	
	delete inputFile;
	
	storeRun = false;
	
	// Initialise integrator
	integrator = new Integrator();
}

ObjectiveFunction::~ObjectiveFunction(){
	//delete rng;
	
	delete integrator;
}

/// Get the fitness of a model. Always specify this function!
double ObjectiveFunction::calculateRawFitness(Model* m){
	/**
 	 * Calculate the fitness from an input-output table
 	 */
	bool steadystate = abortAtSteadyState;	
	if (!m) THROW_ERROR("calculateFitness() called with NULL as argument");

	/* Check if the model contains species and reactions
	* Otherwise, it is useless, and the following routines
	* will probably fail, so return a bad fitness value.
	*/
	if((m->getNumReactions() == 0) || (m->getNumSpecies() == 0)){
		return VERY_BAD_FITNESS;
	}

	double fitness = 0;

	// Set resource to 1.0 (if it is there)
	Species* sp = NULL;
	sp = m->getSpecies("resource");
	if(sp != NULL)	sp->setInitialConcentration(1.0);

	// Now generate a random order in which the cases should be tested
	vector<int> perm;
	perm.reserve(numOfCases);
	for(int i=0; i < numOfCases; i++){
		perm.push_back(i);
	}
	int tmp;
	int rand_pos;
	// Disabled for the moment to ensure a deterministic fitness	
	/*for(int i=0; i < (numOfCases -1); i++){
	tmp = perm.at(i);
	rand_pos = rng->uniform(i,(numOfCases-1));
	perm.at(i) = perm.at(rand_pos);
	perm.at(rand_pos) = tmp;
}
	*/

	vector<integrator_result*> result(numOfCases);
	// getting cvode settings from the objective function
	int printstep = numOfTimesteps-1; // -1 because ODESolver doesn't count 0th step
	double time = (timestep)*printstep; // 
	if(printstep == 0) THROW_ERROR("Only one timepoint given in the objective function, two are needed!");

	// Calculate the timecourse for each case
	for(int currentCase=0; currentCase < numOfCases; currentCase++){
		result.at(currentCase) = NULL;
		// Set the initial concentrations of the input species
		setInitialInputConcentrations(m, perm.at(currentCase), 0);		
// 		SaveSBMLModel(m,"testcase.xml");	
// 		if(currentCase == 2) exit(0);
		// Check if all initial concentrations are only specified for t0
		// In this special case, we can speed up integration a lot (by not stopping it...)
		bool only_t0 = true;
		if(!steadystate){
			for(int i=0; i < numOfCases; i++){
				for(int j=0; j < getNumOfInputSpecies(); j++){
					if(getInitialInputConcentration(i,j,1) >= 0) only_t0 = false;
				}
			}
		}
		else only_t0 = true;

		if(only_t0){
			result.at(currentCase) = integrator->integrateSOSlib(m, time, printstep, steadystate);
			
			// Check for errors
			if(result.at(currentCase) == NULL){
				for(int i=0; i < currentCase; i++){
					delete result.at(i);
				}
				return VERY_BAD_FITNESS;
			}
		}
		else{
			// Integrate the model up to the first timestep
			result.at(currentCase) = integrator->integrateSOSlib(m, timestep, 1);

			// Check for errors
			if(result.at(currentCase) == NULL){
				for(int i=0; i < currentCase; i++){
					delete result.at(i);
				}
				return VERY_BAD_FITNESS;
			}

			// save initial concentrations to restore them later
			vector<double> initConc;
			for(int i=0; i < m->getNumSpecies(); i++){
				initConc.push_back(m->getSpecies(i)->getInitialConcentration());
			}

			// Carry over all concentrations
			for(int i=0; i < m->getNumSpecies(); i++){
				m->getSpecies(i)->setInitialConcentration(result.at(currentCase)->values[m->getSpecies(i)->getId()].at(result.at(currentCase)->timepoints.size()-1));
			}

			// Now loop from timestep to timestep
                	for(int time=2; time <= printstep; time++){
		        // Set the initial concentrations of the input species
				setInitialInputConcentrations(m, perm.at(currentCase), time-1);

				// Integrate the model up to the next timestep
                         	result.at(currentCase) = integrator->integrateSOSlibContinue(m, timestep, 1, result.at(currentCase));

				// Check for errors
				if(result.at(currentCase) == NULL){
					for(int i=0; i < currentCase; i++){
						delete result.at(i);
					}
					return VERY_BAD_FITNESS;
				}	

				// Carry over all concentrations
				for(int i=0; i < m->getNumSpecies(); i++){
					m->getSpecies(i)->setInitialConcentration(result.at(currentCase)->values[m->getSpecies(i)->getId()].at(result.at(currentCase)->timepoints.size()-1));
				}
			}

			// restore saved initial concentrations
			for(int i=0; i < m->getNumSpecies(); i++){
				m->getSpecies(i)->setInitialConcentration(initConc.at(i));
			}
			initConc.clear();
		}
	}

	// Check if run should be stored
	if(storeRun){
		stringstream tmp;
		for(int i=0; i < numOfCases; i++){
			tmp << "case" << i << ".log";
			integrator->saveIntegratorResult(tmp.str().c_str(),result.at(i));
			tmp.str("");
		}
	}
	
	// Now we have all timecourses, get the fitnesses
	// Go over all output species, then over all cases
	for(int species=0; species < getNumOfOutputSpecies(); species++){
		double fitnessHere = 0;
		if(getNameOfOutputSpecies(species) != string("ANY")){
			double diff=0;
			for(int currentCase = 0; currentCase < numOfCases; currentCase++){
				int numTimesteps = result.at(currentCase)->timepoints.size();
				if(numTimesteps != outputConcentrations.at(perm.at(currentCase)).at(species).size()){
					cerr << "Number of timepoints in outputConcentrations and result objects does not match: "
					     << numTimesteps << " vs. " << outputConcentrations.at(perm.at(currentCase)).at(species).size() << endl;
					exit(-1);
				}

				if(numTimesteps <= offset) THROW_ERROR("Due to abortion at steady state, the number of timepoints is smaller than the offset. Maybe delete offset?");					
	
				/* compare the result objective function with the given function F */
				for(int k=offset; k < numTimesteps; k++){
					try{
						diff += pow(outputConcentrations.at(perm.at(currentCase)).at(species).at(k) - result.at(currentCase)->values[getNameOfOutputSpecies(species)].at(k), 2);		
// 						cout << outputConcentrations.at(perm.at(currentCase)).at(species).at(k) << "\t" << result.at(currentCase)->values[getNameOfOutputSpecies(species)].at(k) << endl;
					}
					catch(...){
						cerr << "Something not found in results objects:" << endl
						     << " Output species " << getNameOfOutputSpecies(species) << endl
						     << " Case " << currentCase<< endl
						     << " species " << species << endl
						     << " timepoint " << k << endl;
						SaveSBMLModel(m, "error.xml");
						exit(-1);
					}
				}
			}
			fitnessHere += diff;
		}
		else{
			// Find best species
			double tmp;
			double best = VERY_BAD_FITNESS;
			// Loop over species in results
			for(int m=0; m < result.at(0)->values.size(); m++){
				tmp = 0;

				// Loop over cases
				for(int currentCase = 0; currentCase < numOfCases; currentCase++){
					int numTimesteps = result.at(currentCase)->timepoints.size();
					if(numTimesteps <= offset) THROW_ERROR("Due to abortion at steady state, the number of timepoints is smaller than the offset. Maybe delete offset?");
					for(int k=offset; k < numTimesteps; k++){
						tmp += pow(outputConcentrations.at(perm.at(currentCase)).at(species).at(k) - result.at(currentCase)->values[result.at(currentCase)->ids.at(m)].at(k), 2);
					}
				}
				if(tmp < best){
					best = tmp;
				}
			}	
			fitnessHere += best;
		}
	
		fitness += fitnessHere;
	}

	for(int i=0; i < numOfCases; i++){
		if(result.at(i)) delete result.at(i);
	}

	fitness /= getNumOfOutputSpecies();
	fitness /= numOfCases;
	
	/* It has happend that models have been integrated by CVODE with a very large
	* numerical error, leading to a huge fitness value. However, CVODE has not thrown
	* an error here, thus the model's fitness is not automatically VERY_BAD_FITNESS.
	 * Therefore, we demand that no model can have fitness worse than VERY_BAD_FITNESS/
	* 
	* Models that can conserve mass should always be better than non-mass-conserving ones
	*/
	if(fitness > VERY_BAD_FITNESS) fitness = VERY_BAD_FITNESS -1; 
	return fitness;
}


double ObjectiveFunction::getInitialInputConcentration(int i, int j, int time){
	try{
		return initialInputConcentrations.at(i).at(j).at(time);
	}
	catch(...){
		THROW_ERROR("Input concentration does not exist!");
	}
}

/* 
 * Set concentrations of the input species in a specified case
 */
void ObjectiveFunction::setInitialInputConcentrations(Model* m, int caseNum, int time){
	for(int i=0; i < getNumOfInputSpecies(); i++){
		string id = getNameOfInputSpecies(i);
		if((id.size() > 9) && (id.substr(0,9) == string("CONSTANT_"))){
			Parameter* par = m->getParameter(getNameOfInputSpecies(i).substr(9));
			if(par != NULL){
				if(getInitialInputConcentration(caseNum, i, time) >= 0)	// else means "don't care"
					par->setValue(getInitialInputConcentration(caseNum, i, time));
			}
			else{
				cerr << getNameOfInputSpecies(i) << endl;
				THROW_ERROR("Input parameter (global) not found!");
			}
		}
		else{
			Species* sp = m->getSpecies(getNameOfInputSpecies(i));
			if(sp != NULL){
				if(getInitialInputConcentration(caseNum, i, time) >= 0)	// else means "don't care"
					sp->setInitialConcentration(getInitialInputConcentration(caseNum, i, time));
			}
			else{
				cerr << getNameOfInputSpecies(i) << endl;
				THROW_ERROR("Input species not found!");
			}
		}
	}
	
	// Set output to 0
	/*for(int i=0; i < getNumOfOutputSpecies(); i++){
	Species* sp = m->getSpecies(getNameOfOutputSpecies(i));
	if(sp != NULL){
	sp->setInitialConcentration(0.0);
}
	else{
	THROW_ERROR("Output species not found!");
}
}*/
}



/*
 * Print the values of the objective function to the given ostream
 */
void ObjectiveFunction::print(ostream &outstream){
	if(useAIC) outstream << "Use AIC" << endl;
	else outstream << "Don't use AIC" << endl;
	
	outstream << "Non-massconservance penalty: " << nonMassConservancePenalty << endl;

	outstream << "Size penalty: " << sizePenalty << endl;
	
	for(int numCase=0; numCase < numOfCases; numCase++){
		outstream << "Case no. " << numCase << ":" << endl;
		
		outstream << "Time ";
		for(int i=0; i < getNumOfInputSpecies(); i++){
			outstream << inputSpecies[i] << "\t";
		}

		for(int i=0; i < getNumOfOutputSpecies(); i++){
			outstream << outputSpecies[i] << "\t";
		}
		outstream << endl;
		
		for(int t=0; t < numOfTimesteps; t++){
			outstream << t*timestep << "\t";
			for(int i=0; i < getNumOfInputSpecies(); i++){
				outstream << initialInputConcentrations.at(numCase).at(i).at(t) << "\t";
			}
			for(int i=0; i < getNumOfOutputSpecies(); i++){
				outstream << outputConcentrations.at(numCase).at(i).at(t) << "\t";
			}
			outstream << endl;
		}
	}	
}

#define NEXT_INPUT 	while((inputFile->peek() == '#') || (inputFile->peek() == '\n')) inputFile->ignore(10000, '\n');

void ObjectiveFunction::readInput(istream* inputFile){
	string string_tmp;
	
	useAIC = false;
	
	nonMassConservancePenalty = 10000;
	
	sizePenalty = 0;
	
	// defaults
	numOfCases = 0;
	numOfTimesteps = 0;
	offset = 0;
	timestep = 0;
	
	NEXT_INPUT
	
	string objective_type;		
	// Get type of objective function
	getline(*inputFile, objective_type);
	if(objective_type == string("io-table")) abortAtSteadyState = false;
	else if(objective_type == string("io-table-ss")) abortAtSteadyState = true;
	else THROW_ERROR("Wrong type of objective function file given!"); 


	NEXT_INPUT
			
	// Check whether we use AIC and what the penalty for non-massconservance is
	getline(*inputFile, string_tmp);
	if(string_tmp == string("AIC")) useAIC = true;
	else if(string_tmp == string("noAIC")) useAIC = false;
	else THROW_ERROR("Could not determine whether to use AIC or not from the objective function!");
	
	NEXT_INPUT
	
	getline(*inputFile, string_tmp);
	nonMassConservancePenalty = atof(string_tmp.c_str());
	
	NEXT_INPUT

	// check what the penalty for size is
	getline(*inputFile, string_tmp);
	sizePenalty = atof(string_tmp.c_str());
	
	NEXT_INPUT
			
	// Get number of cases, input species, output species, timesteps, and the timestep
	getline(*inputFile, string_tmp);
	numOfCases = atoi(string_tmp.c_str());
	getline(*inputFile, string_tmp);
	int numOfInputSpecies = atoi(string_tmp.c_str());
	getline(*inputFile, string_tmp);
	int numOfOutputSpecies = atoi(string_tmp.c_str());
	getline(*inputFile, string_tmp);
	numOfTimesteps = atoi(string_tmp.c_str());
	getline(*inputFile, string_tmp);
	offset = atoi(string_tmp.c_str());
	getline(*inputFile, string_tmp);
	timestep = atof(string_tmp.c_str());
	
	if(offset >= numOfTimesteps){
		cout << "Offset: " << offset << " Timesteps: " << numOfTimesteps << endl;
		THROW_ERROR("Offset too large! At least one timestep must be evaluated.");
	}
	
	NEXT_INPUT
	
	// Now read the names of the input species
	for(int i=0; i < numOfInputSpecies; i++){
		getline(*inputFile, string_tmp);
		inputSpecies.push_back(string(string_tmp));
	}
			
	NEXT_INPUT
	
	// Now read the names of the output species
	for(int i=0; i < numOfOutputSpecies; i++){
		getline(*inputFile, string_tmp);
		outputSpecies.push_back(string(string_tmp));
	}
	
	// Check if species is input and output, this might lead to problems
	for(int i=0; i < numOfInputSpecies; i++){
		for(int j=0; j < numOfOutputSpecies; j++){
			if(inputSpecies[i] == outputSpecies[j]) THROW_WARN("An input species is also an output species. This might lead to problems and is not advisable!");
		}
	}
	
	NEXT_INPUT
	
	double tmp_dbl;
	string test_str;


	try{
		// Now read the concentrations for the input species
		for(int i=0; i < numOfCases; i++){
			initialInputConcentrations.push_back(vector< vector <double> >());
			for(int j=0; j < numOfInputSpecies; j++){
				initialInputConcentrations.at(i).push_back(vector<double>());
				NEXT_INPUT
				
				// Read first data point
				// Check if there is data left
				while((inputFile->peek() == ' ') || (inputFile->peek() == '\t')) inputFile->ignore();
				if(inputFile->peek() == '\n'){
					cerr << "Expected more input data points for case " << i << " species " << j << endl;
					THROW_ERROR("Error while reading input data");
				}
				
				// Check if the first position is a '*', which means only the concentration for t0 is to be set
				if(inputFile->peek() == '*'){
					inputFile->ignore();
					*inputFile >> tmp_dbl;
					initialInputConcentrations.at(i).at(j).push_back(tmp_dbl); // '*' encodes that only t0 is to be set
					while((inputFile->peek() == ' ') || (inputFile->peek() == '\t')) inputFile->ignore();
					if(inputFile->peek() != '\n'){
						cerr << "Too much data in input case " << i << " species " << j << endl;
						THROW_ERROR("Error while reading input data");
					}
					for(int t=1; t < numOfTimesteps; t++){
						initialInputConcentrations.at(i).at(j).push_back(-1); // -1 marks "don't care"
					}
				}
				else{
					*inputFile >> tmp_dbl;
					initialInputConcentrations.at(i).at(j).push_back(tmp_dbl);
								
					// Check if there is more data, otherwise assume that the concentration stays constant
					while((inputFile->peek() == ' ') || (inputFile->peek() == '\t')) inputFile->ignore();
					if(inputFile->peek() == '\n'){
						for(int t=1; t < numOfTimesteps; t++){
							initialInputConcentrations.at(i).at(j).push_back(tmp_dbl);
						}
					}
					else{
						// If not constant, read the rest now
						for(int t=1; t < numOfTimesteps; t++){
							// Check if there is data left
							while((inputFile->peek() == ' ') || (inputFile->peek() == '\t')) inputFile->ignore();
							if(inputFile->peek() == '\n'){
								cerr << "Expected more input data points for case " << i << " species " << j << endl;
								THROW_ERROR("Error while reading input data");
							}
							*inputFile >> tmp_dbl;
							initialInputConcentrations.at(i).at(j).push_back(tmp_dbl);
						}
						// check that there is no data left
						getline(*inputFile, test_str, '\n');
						if(test_str.find_first_not_of(string(" \t")) != string::npos){
							cerr << "Too much data in input case " << i << " species " << j << endl;
							THROW_ERROR("Error while reading input data");
						}
						//inputFile->ignore(10000, '\n'); // Go to next line
					}
				}
			}
		}
	
	
		//NEXT_INPUT
	
		// Now read the output species concentrations and construct the 3-D vector structure
		for(int i=0; i < numOfCases; i++){
			outputConcentrations.push_back(vector< vector <double> >());
			for(int j=0; j < numOfOutputSpecies; j++){
				outputConcentrations.at(i).push_back(vector<double>());
				NEXT_INPUT
				
				// Read first data point
				// Check if there is data left
				while((inputFile->peek() == ' ') || (inputFile->peek() == '\t')) inputFile->ignore();
				if(inputFile->peek() == '\n'){
					cerr << "Expected more input data points for case " << i << " species " << j << endl;
					THROW_ERROR("Error while reading input data");
				}
				*inputFile >> tmp_dbl;
				outputConcentrations.at(i).at(j).push_back(tmp_dbl);
				
				// Check if there is more data, otherwise assume that the concentration stays constant
				while((inputFile->peek() == ' ') || (inputFile->peek() == '\t')) inputFile->ignore();
				if(inputFile->peek() == '\n'){
					for(int t=1; t < numOfTimesteps; t++){
						outputConcentrations.at(i).at(j).push_back(tmp_dbl);
					}
				}
				else{
					for(int t=1; t < numOfTimesteps; t++){
						// Check if there is data left
						while((inputFile->peek() == ' ') || (inputFile->peek() == '\t')) inputFile->ignore();
						if(inputFile->peek() == '\n'){
							cerr << "Expected more output data points for case " << i << " species " << j << endl;
							THROW_ERROR("Error while reading output data");
						}
						*inputFile >> tmp_dbl;
						outputConcentrations.at(i).at(j).push_back(tmp_dbl);
					}
					// check that there is no data left
					getline(*inputFile, test_str, '\n');
					if(test_str.find_first_not_of(string(" \t")) != string::npos){
						cerr << "Too much data in output case " << i << " species " << j << endl;
						THROW_ERROR("Error while reading output data");
					}
					//inputFile->ignore(10000, '\n'); // Go to next line
				}
			}
		}
		
	} // try
	catch(...){
		THROW_ERROR("An error occured while constructing the objective function!");
	}
}	

