
 
#include "parameterfitter.h"

using namespace std;
using namespace sbmlevolver;


// constructor ParameterFitter
ParameterFitter::ParameterFitter(int lambda, int mu, int numOfGenerations, bool useOldStartingValues, unsigned long int seed):
	rng(seed), 
	lambda(lambda), 
	mu(mu), 
	numOfGenerations(numOfGenerations)
{
	minimalLowerBound = 1E-10;	
	use_old_starting_values = useOldStartingValues;
}


// New version using a modified code from the ANSI-C implementation of CMA-ES by Nikolaus Hansen
// http://www.bionik.tu-berlin.de/user/niko/cmaes_inmatlab.html
// Based on the file example2.c, I have removed the restart mechanism and the file I/O
//
// optimize the not-fixed parameters of the Individual ind
void ParameterFitter::fitParameters (Model* m, ObjectiveFunction* F, bool quiet){

	//check if fitting is wanted
	if (numOfGenerations==0) return;
	
	cmaes_t evo;       /* the optimizer */
	double *const*pop; /* sampled population */
	double *fitvals;   /* objective function values of sampled population */
	double fbestever=0, *xbestever=NULL; /* store best solution */
	double fmean; 
	char const * stop; /* stop message */
	const char* filename = "non"; // this should prevent file input

	//Read all parameter of the Model
	vector<PInfo*> allInfo;		//important to save allInfo
	getIndividualInfo(allInfo, m);
	
	//Select all "mutatable" parameter
	vector<PInfo*> mutatable;
	for (int i = 0; i< allInfo.size(); i++){
		if (! allInfo[i]->isFixed){
			mutatable.push_back(allInfo[i]);
		}
	}
	
	Dimension = mutatable.size();	//dimension

	normalize(mutatable);

	// get array of starting values and starting stddevs
	double* xstart = (double*) malloc(Dimension*sizeof(double));
	double* sdstart = (double*) malloc(Dimension*sizeof(double));
	for(int i=0; i < Dimension; i++){
		xstart[i] = mutatable.at(i)->norm_value;
		//cout << "Par " << i << " val: " << mutatable.at(i)->value << endl;
		sdstart[i] = mutatable.at(i)->initialStepsize;
		//cout << "Par " << i << " " << xstart[i] << " " << sdstart[i] << endl;
	}

	/* Parameters can be set in three ways. Here as input parameter 
	 * to cmaes_init, as value read from initials.par in readpara_init
	 * during initialization, and as value read from signals.par by
	 * calling cmaes_ReadSignals explicitely. 
	 */
	fitvals = cmaes_init(&evo, Dimension, xstart, sdstart, rng.getASeed(), lambda, filename, numOfGenerations); /* allocs fitvals */
	//printf("%s\n", cmaes_SayHello(&evo)); 
	//cmaes_ReadSignals(&evo, "signals.par"); /* write initial values, headers in case */

	free(xstart);
	free(sdstart);
	xstart = NULL;
	sdstart = NULL;
			
	evaluateMe = (Model*) m->clone();	//copy to avoid sideeffects
	
	// loop
	while(!(stop=cmaes_TestForTermination(&evo))){ 
		/* Generate population of new candidate solutions */
		pop = cmaes_SamplePopulation(&evo); /* do not change content of pop */
			
		/* Compute fitness value for each candidate solution */
		for (int i = 0; i < cmaes_Get(&evo, "popsize"); ++i) {
			// OK, I'm not totally sure with this, but I
			// will use sticky bounds, because I expect my
			// optima to be on the boundary quite often...
			// N. Hansen recommends resampling or a penalty approach
			forceIntoBounds(pop[i]);
			fitvals[i] = callFitnessFunction(pop[i],allInfo,F); 
			//for(int k=0; k < Dimension; k++){
			//	cout << pop[i][k] << " ";
			//}
			//cout << ": " << fitvals[i] << endl;
		}
		
		/* update search distribution */
		cmaes_UpdateDistribution(&evo, fitvals); 
		
		/* read control signals for output and termination */
		//cmaes_ReadSignals(&evo, "signals.par"); /* from file signals.par */
		
		// output generation and best fitness
		//cout << cmaes_Get(&evo, "generation") << " " << cmaes_Get(&evo,"fitness") << endl;
		
		fflush(stdout);
	} /* while !cmaes_TestForTermination(&evo) */

	if(!quiet){
		/* print some "final" output */
		printf("%.0f generations, %.0f fevals (%.1f sec): f(x)=%g\n", 
		       cmaes_Get(&evo, "gen"), cmaes_Get(&evo, "eval"), 
		       evo.eigenTimings.totaltime,
		       cmaes_Get(&evo, "funval"));
		printf("  (axis-ratio=%.2e, max/min-stddev=%.2e/%.2e)\n", 
		       cmaes_Get(&evo, "maxaxislen") / cmaes_Get(&evo, "minaxislen"),
		       cmaes_Get(&evo, "maxstddev"), cmaes_Get(&evo, "minstddev") 
		       );
		printf("Stop run:\n%s\n", cmaes_TestForTermination(&evo));
	}

	fbestever = cmaes_Get(&evo, "fbestever"); 
	xbestever = cmaes_GetInto(&evo, "xbestever", xbestever); /* alloc mem if needed */

	/* best estimator for the optimum is xmean, therefore check */
	if ((fmean = callFitnessFunction(cmaes_GetPtr(&evo, "xmean"),allInfo,F)) < fbestever) {
		fbestever = fmean;
		xbestever = cmaes_GetInto(&evo, "xmean", xbestever);
	}
	
	const double *stddev = cmaes_GetPtr(&evo, "stddev");
	//for(int i=0; i < Dimension; i++){
	//	cout << "STDDEV" << i << " " << stddev[i] << endl;
	//}

	setInitialStepsizes(m,allInfo,cmaes_GetPtr(&evo, "stddev"));
	
	updateModelParameter(xbestever, m, allInfo);
	
	//clean up
	for (int i = 0; i < allInfo.size(); i++)
		delete allInfo[i];
	allInfo.clear();	
	
	mutatable.clear();

	free(xbestever);
	xbestever=NULL;
	cmaes_exit(&evo); /* does not effect the content of stop string and xbestever, but frees fitvals and pop */
	
	if (evaluateMe)	delete evaluateMe;
}

void ParameterFitter::forceIntoBounds(double* x){
	for(int i=0; i < Dimension; i++){
		if(x[i] < 0) x[i] = 0;
		if(x[i] > 1) x[i] = 1;
	}
}


//update Model parameter
void ParameterFitter::updateModelParameter(const double* x, Model* m, vector<PInfo*>& allInfo){
	//assign normalized parameter to the parameter infos for the whole model
	int variable = 0; 
	for (int k = 0; k < allInfo.size(); k++){
		if (!allInfo[k]->isFixed){
			allInfo[k]->norm_value=x[variable];
			variable++;
		}
	}
	
	//denormalize Parameter
	denormalize(allInfo);	
	
	//change model
	setModelParameters(m,allInfo);
}


void ParameterFitter::normalize(vector<PInfo*> mySet){
	if(mySet.size() != Dimension){
		THROW_ERROR("Wrong dimension in function normalize!");
	}

	for (int i = 0; i < Dimension; i++){
		//calculation of a
		mySet.at(i)->a = log10(max(minimalLowerBound,mySet.at(i)->minVal));
		
		//calculation of b
		//if upper bound is equal to minimal lower bound or lower you have to set it to more than
		//minimal lower bound to make upper and lower bound unequal (so 2 times minimal lower bound)
		if (mySet.at(i)->maxVal<=minimalLowerBound)
		        mySet.at(i)->b = log10(minimalLowerBound*2) - mySet.at(i)->a;		
		else
			mySet.at(i)->b = log10(mySet.at(i)->maxVal) - mySet.at(i)->a;
		
		//calculation of normalized value
		mySet.at(i)->norm_value = (log10(max(minimalLowerBound,mySet.at(i)->value)) - mySet.at(i)->a)/(mySet.at(i)->b);
	}
}

void ParameterFitter::denormalize(vector<PInfo*> mySet){
	for (int i = 0; i < mySet.size(); i++){
		if(! mySet.at(i)->isFixed){
			//calculation of unnormalized value
			mySet.at(i)->value = pow(10,(mySet.at(i)->a + (mySet.at(i)->b*mySet.at(i)->norm_value)));
			if (mySet.at(i)->value <= minimalLowerBound) mySet.at(i)->value = 0;
		}
	}
}





double ParameterFitter::callFitnessFunction(const double* x, vector<PInfo*>& allInfo, ObjectiveFunction* F){
	int variable = 0; 
	for(int i=0; i < allInfo.size(); i++){
		if(!allInfo.at(i)->isFixed){
			allInfo.at(i)->norm_value = x[variable];
			variable++;
		}
	}
	denormalize(allInfo);
	
	setModelParameters(evaluateMe, allInfo);

	double tempFitness = F->evaluate(evaluateMe);	
	
	return tempFitness;
}


void ParameterFitter::setModelParameters(Model* mod, vector<PInfo*>& indInfo) {
	// the input indInfo is actually allInfo, so all parameters, fixed and non-fixed ones
	int parameterIndex = 0; // runs over all global & local parameters & speciesInitialConc of the Individual

	// 1) change global model-parameters
	Parameter* par;
	for(int i=0; i< mod->getNumParameters(); i++) {
		if(!(indInfo[parameterIndex]->isFixed)){
			par = mod->getParameter(i);
			setParameter(mod->getParameter(i), indInfo[parameterIndex]->value, true);
		}
		parameterIndex++;
	}

	// 2) change local model-parameters
	KineticLaw* kl = 0;
	for(int i=0; i< mod->getNumReactions(); i++) {
		kl = mod->getReaction(i)->getKineticLaw();
		for(int j=0; j< kl->getNumParameters(); j++) {
			if(!(indInfo[parameterIndex]->isFixed)){
				par = kl->getParameter(j);
				setParameter(par, indInfo[parameterIndex]->value, true);
			}
			parameterIndex++;
		}		
	}

	// 3) change species initial concentration
	for(int i=0; i<mod->getNumSpecies() ; i++){
		if(!(indInfo[parameterIndex]->isFixed)){
			setSpeciesInitialConcentration(mod->getSpecies(i), indInfo[parameterIndex]->value, true);
		}
		parameterIndex++;
	}
}


void ParameterFitter::setInitialStepsizes(Model* mod, vector<PInfo*>& indInfo, const double* stepSizes) {
	int variableIndex = 0; // runs over all components that are changed by the parameter fitter
	int parameterIndex = 0; // runs over all global & local parameters & speciesInitialConc of the Individual

	// 1) change global model-parameters
	Parameter* par;
	for(int i=0; i< mod->getNumParameters(); i++) {
		if(!(indInfo[parameterIndex]->isFixed)){
			par = mod->getParameter(i);
			setAttr(par,"initial_stepsize",stepSizes[variableIndex]);
			variableIndex++;
		}
		parameterIndex++;
	}

	// 2) change local model-parameters
	KineticLaw* kl = 0;
	for(int i=0; i< mod->getNumReactions(); i++) {
		kl = mod->getReaction(i)->getKineticLaw();
		for(int j=0; j < kl->getNumParameters(); j++) {
			if(!(indInfo[parameterIndex]->isFixed)){
				par = kl->getParameter(j);
				setAttr(par,"initial_stepsize",stepSizes[variableIndex]);
				variableIndex++;
			}
			parameterIndex++;
		}		
	}

	// 3) change species initial concentration
	for(int i=0; i<mod->getNumSpecies() ; i++){
		if(!(indInfo[parameterIndex]->isFixed)){
			setAttr(mod->getSpecies(i),"initial_stepsize",stepSizes[variableIndex]);
			variableIndex++;
		}
		parameterIndex++;
	}
}


// Check which parameters are fixed & get Info about all parameters of Individual
// fill the vector indInfo by push_back(PInfo*);
void ParameterFitter::getIndividualInfo(vector<PInfo*>& indInfo, Model* m) {
	int parameterIndex = 0; // count over all parameters of ind
	
	// 1) global model-parameters
	for(int i=0; i< m->getNumParameters(); i++) {
		Parameter* p = m->getParameter(i);
		indInfo.push_back(parseParameter(p));
		parameterIndex++;
	}
	
	// 2) local parameters from kinetic laws
	for(int i=0; i< m->getNumReactions(); i++) {
		KineticLaw* kl = m->getReaction(i)->getKineticLaw();
		for(int j=0; j< kl->getNumParameters(); j++) {
			Parameter* p = kl->getParameter(j);
			indInfo.push_back(parseParameter(p));
			parameterIndex++;
		}		
	}
	
	// 3) species initial concentrations
	for(int i=0; i < m->getNumSpecies(); i++){
		Species* s = m->getSpecies(i);
		indInfo.push_back(parseSpecies(s));
		parameterIndex++;
	}
}

// get information about a (global or kineticlaw) model parameter:
// 1) it is fixed or not (get min & max val if not)
// 2) it's value ist set or not (get value if not)
// returns Pointer to PInfo, containing this information
ParameterFitter::PInfo* ParameterFitter::parseParameter(Parameter* p) {
	PInfo* pInfo = new PInfo();
	string pName = p->getName();
	if(isMutational(p)){
		if((!use_old_starting_values) && (getAttr(p,"lower_bound") < getAttr(p,"upper_bound"))){
			p->unsetValue();
		}
		else{
			// Check that parameter is inside boundary
			if(p->getValue() < getAttr(p,"lower_bound")) p->setValue(getAttr(p,"lower_bound"));
			if(p->getValue() > getAttr(p,"upper_bound")) p->setValue(getAttr(p,"upper_bound"));
		}
		if(!isSetAttr(p,"initial_stepsize")) setAttr(p,"initial_stepsize",0.1);
		if(getAttr(p,"initial_stepsize") == 0) setAttr(p,"initial_stepsize",0.1);
	}	

	
	if( p->isSetValue() ) {					     // get its value if set
		pInfo->value = p->getValue();
	} else {
		pInfo->value = getAttr(p,"lower_bound") + rng.uniform()*(getAttr(p,"upper_bound")-getAttr(p,"lower_bound"));
	}
	
	if((! isMutational(p)) || (getAttr(p,"lower_bound") >= getAttr(p,"upper_bound")))  {  // parameter is constant
		pInfo->isFixed = true;
	} else {									 // parameter isn't constant -->
		pInfo->minVal = getAttr(p,"lower_bound");
		pInfo->maxVal = getAttr(p,"upper_bound");
		pInfo->initialStepsize = getAttr(p,"initial_stepsize");
		pInfo->isFixed = false;
		
		// check that parameter is inside boundaries
		if(pInfo->value < pInfo->minVal) pInfo->value = pInfo->minVal;
		if(pInfo->value > pInfo->maxVal) pInfo->value = pInfo->maxVal;
	}
	
	//cout << p->getName() << " " << p->getValue() << " " << pInfo->value << " " << pInfo->minVal << " " << pInfo->maxVal << endl;
	return pInfo;
}

ParameterFitter::PInfo::PInfo(){
	minVal = 0;
	maxVal = 0;
	value = 0;
	a = 0;
	b = 0;
	norm_value=0;
	isFixed = false;
	initialStepsize = 0.01;
}

// get information about the initial concentration of a Species:
// 1) it is fixed or not (get min & max val if not)
// 2) it's value ist set or not (get value if not)
// returns Pointer to PInfo, containing this information
ParameterFitter::PInfo* ParameterFitter::parseSpecies(Species* s) {
	PInfo* pInfo = new PInfo();
	string pName = s->getName();

	if((!use_old_starting_values) && (getAttr(s,"lower_bound") < getAttr(s,"upper_bound"))){
		s->unsetInitialConcentration();
	}
	else{
		// check that initial concentration is inside bounds
		if(s->getInitialConcentration() < getAttr(s,"lower_bound")) s->setInitialConcentration(getAttr(s,"lower_bound"));
		if(s->getInitialConcentration() > getAttr(s,"upper_bound")) s->setInitialConcentration(getAttr(s,"upper_bound"));
	}
	if(!isSetAttr(s,"initial_stepsize")) setAttr(s,"initial_stepsize",0.1);
	if(getAttr(s,"initial_stepsize") == 0) setAttr(s,"initial_stepsize",0.1);

	
	if( s->isSetInitialConcentration() ) {		 // get its value if set
		pInfo->value = s->getInitialConcentration();
	} else {
		pInfo->value = getAttr(s,"lower_bound") + rng.uniform()*(getAttr(s,"upper_bound")-getAttr(s,"lower_bound"));
	}
	
	pInfo->minVal = getAttr(s,"lower_bound");
	pInfo->maxVal = getAttr(s,"upper_bound");
	pInfo->initialStepsize = getAttr(s,"initial_stepsize");
		
	// check that parameter is inside boundaries
	if(pInfo->value < pInfo->minVal) pInfo->value = pInfo->minVal;
	if(pInfo->value > pInfo->maxVal) pInfo->value = pInfo->maxVal;
	
	// check if its fixed
	if(pInfo->minVal < pInfo->maxVal) pInfo->isFixed = false;
	else pInfo->isFixed = true;
	
	return pInfo;
}


void ParameterFitter::setParameter(Parameter* para, double val, bool override){
	if(override || (! para->isSetValue())){
		para->setValue(val);
	}
}

void ParameterFitter::setSpeciesInitialConcentration(Species* sp, double val, bool override){
	if(override || (! sp->isSetInitialConcentration())){
		sp->setInitialConcentration(val);
	}
}

