/************************************************************************
  			parameterfitter.h - SBMLEvolver - CSB, FSU Jena

This software is free for use and modification, but comes without any
warranty. Usage is entirely at your own risk!

Developers: 	Thorsten Lenser (thlenser@inf.uni-jena.de)
		Lisa Hellwig (louis-chan@web.de)

Fit the parameters of the model, using an evolutionary strategy approach
described in Runarsson & Yao 2000

	$Id: parameterfitter.h 28 2007-06-14 07:55:32Z thlenser $

**************************************************************************/
 

#ifndef PARAMETERFITTER_H
#define PARAMETERFITTER_H

#include <string>
#include <vector>
#include <cmath>
#include <algorithm>
#include <cstdlib>
#include <sstream>

#include "cmaes_interface.h"

#include "include.h"
#include "objectivefunction.h"
#include "randomnumbergenerator.h"
#include "sbmltools.h"

using namespace std;

namespace sbmlevolver{

/**
 *\brief Fits parameters
 *
 * Parameterfitter with CMA-ES strategy
 */
class ParameterFitter {

public:
	/// Constructor
	ParameterFitter(int lambda, int mu, int numOfGenerations, bool useOldStartingValues, unsigned long int seed);

	/// Fit the parameters of the model to match the objective function
	void fitParameters (Model* m, ObjectiveFunction* F, bool quiet);
	
private:
	/// model for evaluation
	Model* evaluateMe;
	/// Random number generator
	RandomNumberGenerator rng;
	/// Size of population (lambda)
	const int lambda;
	/// Number of parents (mu), must be smaller than size of population
	const int mu;	
	/// Number of generations the fitting should run
	const int numOfGenerations;
	/// Number of parameter that have to be fitted
	int Dimension;
	/// decide whether the old parameter values should be used as starting values (true) or new ones should be generated (false)
	bool use_old_starting_values;
	double minimalLowerBound;

	/// Information about parameter
	struct PInfo {
		/// minimal value
		double minVal;
		/// maximal value
		double maxVal;
		/// value
		double value;
		/// constants for normalisation
		double a;
		double b;
		/// normalised value (between 0 and 1)
		double norm_value;

		/// true if the parameter is fixed and should not be fitted
		bool isFixed;
		/// initial stepsize
		double initialStepsize;
		
		/// Constructor
		PInfo();
	};
	
	 void normalize(vector<PInfo*>);
	 void denormalize(vector<PInfo*>);

	 void forceIntoBounds(double*);

	 // Calls fitness function F for model m with new calculated parameters in indInfo
	 // - uses fixed parameters in allInfo
	 double callFitnessFunction(const double* x, vector<PInfo*>& allInfo, ObjectiveFunction* F);

	 //denormalizes parameters and calls setModelParameterCMA
	 void updateModelParameter(const double* x, Model* m, vector<PInfo*>& allInfo);

	 // change all parameters of 'mod' into the parameters of 'indInfo'
	 void setModelParameters(Model* mod, vector<PInfo*>& indInfo);

	 // change stepsize of all mutatable parameters to tmpStepSize 
	 void setInitialStepsizes(Model* mod, vector<PInfo*>& indInfo, const double* stepSizes);

	 /// Check which parameters are fixed
	 void getIndividualInfo(vector<PInfo*>& indInfo, Model* m);

	 // get information about a (global or kineticlaw) model parameter:
	 PInfo* parseParameter(Parameter* p);

	 /// Get information for SpeciesInitialConcentration (see parseParameter)
	 PInfo* parseSpecies(Species* s);

	 /// Set the parameter to the given value, override old value if that is specified
	 void setParameter(Parameter*, double, bool);

	 /// Set the initial concentration of the species, override old value if specified
	 void setSpeciesInitialConcentration(Species*, double, bool);
};

} // namespace sbmlevolver

#endif //PARAMETERFITTER_H

