/***********************************************************************
 *       paramobj.h - SBMLEvolver - CSB, FSU Jena
 * 
 * This software is free for use and modification, but comes without any
 * warranty. Usage is entirely at your own risk!
 *
 * This class capsulates all parameters given to the SBMLevolver
 * 
 *	THREAD SAFETY
 *	  nothing is reentrant, except all createXXX()-functions
 *
 *	$Id: paramobj.h 29 2007-06-15 13:35:45Z thlenser $
 *	  
 ************************************************************************/
#ifndef PARAM_H
#define PARAM_H

#include <vector>
#include <time.h>
#include <string.h>
#include <sstream>
#include <fstream>

#include "include.h"

namespace sbmlevolver {

/**
 *\brief Manages SBMLevolver parameters
 *
 * This class capsulates all parameters given to the SBMLevolver.
 * It also holds a set of objects of which only one copy is needed
 * in the program, giving all program components central access to 
 * these. 
 */
class ParamObj {
	public:
	/// Default constructor
	ParamObj();
	/// constructor with optionsfile (use this usually!)
	ParamObj(string optionsfile);
	
	/// Destructor
	~ParamObj();

	/// parse options 
	void parseOptions(istream* input);
	/// parse options file
	void parseOptionsFile(string filename);
	/// Write object to outstream
	void print(ostream &outstream);
	/// Print help information
	void printHelp(ostream &outstream);

	/* parameter record forward defintions */
	/// parameters for overall session
	struct ParamSession;
	/// parameters for population
	struct ParamPopulation;
	/// parameters for mutator
	struct ParamMutator;
	/// parameters for parameter fitter
	struct ParamParaFitter;

	/// parameters for mutator
	struct ParamMutator {
		/// Probabilities for the different mutations
		vector<double> mutationProbs; 
		/// maximal number of mutations in one step
		unsigned int maxSimultan;
	} mutator;


	/* the params itself */
	// session (not portable to other client or next run)
	/// parameters for overall session
	struct ParamSession {
		// I/O
		/// additional output file for machine readable output
		string outputFile;
		// working dir before running the session
		//string inputDir;
		/// working while and after run
		string outputDir;	
		
		// program termination
		/// fitness to stop at (smaller or equal)
		double stopFit;		
		/// number of generations to pass before stop
		unsigned long stopTurn;	
		/// unix time to stop at
		time_t stopTime;	
		
		/// Save some individual every x generations
		int saveEvery;
		/// How many should we save
		int numSave;
		
	} session;
	
	/// parameters for population
	struct ParamPopulation {
		/// number of populations
		unsigned int num_populations;
		/// migrate best individuals every x generations
		int isolationTime;
		/// size of population (mu)
		unsigned int mu;
		/// number of offspring (lambda)
		unsigned int lambda;
		/// filename of file containing initial starting individual
		string firstModelFile;
		/// Probability of crossover (= 1-probability of mutation)
		double crossover_prob;
		/// Selection scheme, comma or plus
		bool comma;
		/// Always re-evaluate fitness function at each generation
		bool alwaysEvaluate;
		/// Number of species in each model in initial population
		int initialNumSpecies;
		/// Number of reactions in each model in initial population
		int initialNumReactions;
	} population;
	
	/// parameters for parameter fitter
	struct ParamParaFitter {
		/// population size (lambda)
		int populationSize;
		/// number of parents (mu)
		int numOfParents;
		/// number of generations used for fitting
		int numOfGenerations;
		/// use old values for starting or not
		bool useOldStartingValues;
	} paraFitter;


	/// Seed for random number generator. All other seeds should be generated from this one!
	unsigned long int seed;

	/// stores the buildingblocks file
	string BBstring; 
	/// stores the objective function file
	string OFstring; 
	
	
protected:
	/// initialise default parameters
	void initDefaults();
};

} // namespace

#endif // PARAM_H
