/******************************************************************************
  			population.h - SBMLEvolver - CSB, FSU Jena

This software is free for use and modification, but comes without any
warranty. Usage is entirely at your own risk!

Developers: 	Thorsten Lenser (thlenser@inf.uni-jena.de)
		Jan Huwald (jh@sotun.de)

	Population holds a population, implements selection,
  	cares for generations (e.g. create offspring), manages a few stats
	and holds an interface to share the work

	    
	$Id: population.h 3 2007-06-04 12:59:46Z thlenser $
	    
******************************************************************************/
 

#ifndef POPULATION_H
#define POPULATION_H

#include <string>
#include <map>
#include <set>
#include <deque>
#include <cmath>
#include <dirent.h>

#include "include.h"
#include "individual.h"
#include "randomnumbergenerator.h"
#include "paramobj.h"
#include "mutator.h"

using namespace std;

namespace sbmlevolver{
	
/**
 * \brief Manages the population
 *
 * Population holds a population, implements selection, cares for generations
 * (e.g. create offspring), manages a few stats and holds an interface to share the work
 *
 * population dynamics / selection:
 * currently a centralized, generation based approach is implented:
 * after determining the fitness of all individuals currently hold all
 * but the (numSurvivers) best are deleted; from them offspring is
 * generated until (PopSize) number of individuals are reached.
 *
 * population creation:
 * a (FirstModel) model is given to the constructor
 * every time the gene pool is empty, it is filled up with mutations of this model
 */	
class Population {

public:
	/// Constructor
	Population(unsigned long int seed, ParamObj* params);
	/// Destructor
	~Population();
	
	// simulation
	/// incremented the current generation number
	void incCurrentGeneration();	
	/// return the current generation number
 	IndGen getCurrentGeneration();
	/// Selection
	void select();
	/// Offspring creation
	void createOffspring();
	/// Calculates the fitnesses for unrated individuals in a single thread
	void calculateFitnessValues(); 
	
	/// unset fitness values for the whole population
	void unsetFitnessOfWholePopulation();

	// Kernel-based fitness sharing which helps to maintain diversity
	//void fitnessSharing();
	
	// population & selection parameters
	/// get mu
	int getMu();
	/// get lambda
	int getLambda();

	// population stats & managment
	/// return a set of the best numSave individuals
	set<Individual*> *getBestIndividuals(int numSave); // (extern only) function to get all models of all ind's
	/// get current number of individuals in population
	int getNumOfIndividuals();
	/// get fitness of best individual
	double getBestFitness();
	/// get fitness of worst individual
	double getWorstFitness();
	/// get average fitness of current population
	double getAverageFitness();
	/// get variance of fitnesses
	double getVariance();
	/// get the average structural difference
	double getAverageStructuralDifference();
	/// get the average structural difference
	double getAverageStructuralDifference2();
	/// get the average number of species
	double getAverageNumSpecies();
	/// get the average number of reactions
	double getAverageNumReactions();
	/// get a vector of all fitnesses
	vector<double> getAllFitnesses();
	
	// individuals
	/// add an individual to the population, without changing its ID and generation number
	void addIndividual (Individual* ind); // create individual from model (m) and insert w/ no regards to existing borders (e.g. pop-size)
	/// add an individual which is to be loaded from a model-file
	void loadIndividualFromFile(char* filename); 
	/// add an initial solution, created by mutations from first mother
	void addInitialSolution();	
	
	// Island strategy
	/// add an individual from another population, set its ID and generation number
	void receiveIndividual(Individual* ind);
	/// get a copy of best individual with ID and generation set to 0
	Individual* extractBestIndividual();
	
	/// get an unevaluated individual and remove it from the population
	Individual* extractTask(); 
	/// thread could not solve problem and returns it to problem pool
	void cancelTask(Individual *ind); 
	/// submit the evaluated individual back into the population
	void submitResult(Individual *ind);
	/// return the number of task which are currently being work on outside the population
	int tasksInProgress();
		
protected:

	// selection & offspring
	/// population size
	unsigned int mu;
	/// number of offspring
	unsigned int lambda;
	/// type of selection scheme, comma or plus
	bool comma;
	/// density pressure for population diversity
	double alpha;

	/// probability for crossover
	double crossover_prob;


	/// the "seed" from which the population is created (usually minimal SBML model)
	Individual *FirstModel;
	/// random number generation
	RandomNumberGenerator *rng;
	/// Building Blocks object
	BuildingBlocks* bb;
	/// Objective Function object
	ObjectiveFunction* of;
	/// Mutator object
	Mutator* mut;

	
	/// select Ind randomly (equal propability for all)
	Individual *RandomSelectInd();		
	// select Ind weight (positivly) by fitness (->most likely the fittest)
	// Individual *RandomSelectInd_WeightFit(); // not implemented yet	
	/// select individual with probability proportional to fitness (we are minimising fitness)
	Individual *RandomSelectInd_WeightUnfit(int);	
	
	// individuals
	/// remove given individual from population
	void removeIndividual(Individual *ind);
	
	// individuals (& their indexes)
	/// individuals listed by fitness (individuals without fitness are not in here)
	multimap<double, Individual*> inds;
	/// individuals with currently unknown fitness
	deque<Individual*> openTasks;	
	
	// population statistics
	/// the currently highest ID, also the number of created individuals
	IndId HighestId;
	/// the current generation
	IndGen CurrentGeneration;
	
	/// number of threads currently working on a problem (can change at all times!) 
	int openTasksOutsidePop; 
};

} // namespace sbmlevolver

#endif //POPULATION_H

