/************************************************************************
		RandomNumberGenerator.h - SBMLEvolver - CSB, FSU Jena

This software is free for use and modification, but comes without any
warranty. Usage is entirely at your own risk!

Developers: 	Thorsten Lenser (thlenser@inf.uni-jena.de)

Declaration of the class RandomNumberGenerator, which provides
methods generating random numbers following specific distributions.
This is meant to capsule a third-party RNG, so that we can easily
exchange the RNG if problems arise.

The random number generator uses the GNU Scientific Library (GSL),
which has to be installed separately. Currently, the "Mersenne Twister"
by Makoto Matsumoto and Takuji Nishimura is used.
See http://www.gnu.org/software/gsl/manual/html_node/index.html

	$Id: randomnumbergenerator.h 3 2007-06-04 12:59:46Z thlenser $

**************************************************************************/

#ifndef RANDOMNUMBERGENERATOR_H_
#define RANDOMNUMBERGENERATOR_H_

#include <cstdlib>
#include <ctime>

#include "gsl/gsl_rng.h"
#include "gsl/gsl_randist.h"

#include "include.h"

using namespace std;

namespace sbmlevolver{

/**
 * \brief Provides pseudo-random numbers
 *
 * Class RandomNumberGenerator provides	methods generating random numbers 
 * following specific distributions. This is meant to encapsulate a 
 * third-party RNG, so that we can easily exchange the RNG if problems arise.
 * 
 * The random number generator uses the GNU Scientific Library (GSL),
 * which has to be installed separately. Currently, the "Mersenne Twister"
 * by Makoto Matsumoto and Takuji Nishimura is used.
 * See http://www.gnu.org/software/gsl/manual/html_node/index.html
 */
class RandomNumberGenerator
{
private:
	/// Internal random number generator from GSL
	gsl_rng* r;
	
public:
	// Constructor using random seed
	//RandomNumberGenerator();
	/// Constructor using user-defined seed
	RandomNumberGenerator(unsigned long int seed);
	/// Destructor
	virtual ~RandomNumberGenerator();
	
	/// Provides a uniform integer random number in the range min,..,max
	int uniform(int, int);
	
	/// Provides a uniform random number in the range [0..1)
	double uniform();
	
	/// Provides a gaussian random number with given mean and standard deviation
	double gaussian(double, double);
	
	/// Takes a double 0 <= x <= 1, and returns true with that probability
	bool challenge(double);

	unsigned long int getASeed();

};

} // namespace sbmlevolver

#endif /*RANDOMNUMBERGENERATOR_H_*/
