/* Version information
 *
 * $Id: sbmltools.cpp 30 2007-06-15 15:32:33Z thlenser $
 *
 */


#include "sbmltools.h"

using namespace std;
namespace sbmlevolver{
	

/**
 * Replace sp1 by sp2
 */
void replaceSpeciesInReaction(Reaction* re, string sp1, string sp2){
	if(sp1 == sp2) return;

	// Replace reactant species reference
	SpeciesReference* sr = re->getReactant(sp1);
	while(sr != NULL){
		 sr->setSpecies(sp2);
		 sr = re->getReactant(sp1);
	}
	
	// Replace product species reference
	sr = re->getProduct(sp1);
	while(sr != NULL){
		sr->setSpecies(sp2);
		sr = re->getProduct(sp1);
	}
	
	// Replace modifier species reference
	ModifierSpeciesReference* msr = re->getModifier(sp1);
	while(msr != NULL){
		msr->setSpecies(sp2);
		msr = re->getModifier(sp1);
	}

	// Replace species in kinetic law
	string formula;
	if(re->isSetKineticLaw()){
		if(re->getKineticLaw()->isSetFormula()){
			formula = re->getKineticLaw()->getFormula();
		}
		else{
			THROW_ERROR("Found kinetic law without math!");
		}
	}
	else{
		THROW_ERROR("Found reaction without kinetic law!");
	}
	string::size_type index;
	index = formula.find(sp1,0);
	while(index != string::npos){
		// Check that found pattern is not just part of another (longer) species id
		if((index > 0) && (isalnum(formula.at(index-1)))){
			// do nothing
		}
		else if(((index + sp1.size()) < formula.size()) && (isalnum(formula.at(index + sp1.size())))){	// name continues
			// move index to end of name
			index += sp1.size();
			while((index < formula.size()) && isalnum(formula.at(index))) index++;
		}
		else{
			formula.erase(index, sp1.length());
			formula.insert(index, sp2);
		}
		if(index < (formula.size() -1)) index = formula.find(sp1, index+1);  // Avoid cycle when replacing (e.g.) X1 by X10
		else index = string::npos;
	}
	
	re->getKineticLaw()->setFormula(formula);
}

// Little helper
int factorial (int n) {
	int f = 1;

	for (int i=1; i <= n; i++)
		f = f*i;
	return f;
}

/**
 * The two reactions are equal if they have the same reactants, modifiers and products
 * and the same kinetic law
 */
bool equal(Reaction* re1, Reaction* re2){
	// Check that reactants, modifiers and products fit
	// try every permutation
	int numReactants = re1->getNumReactants();
	int numModifiers = re1->getNumModifiers();
	int numProducts = re1->getNumProducts();
	
	if(numReactants != re2->getNumReactants()) return false;
	if(numModifiers != re2->getNumModifiers()) return false;
	if(numProducts != re2->getNumProducts()) return false;
	if(re1->getKineticLaw()->getNumParameters() != re2->getKineticLaw()->getNumParameters()) return false;
	
	vector<int> perm;
	perm.reserve(numReactants);
	for(int j=0; j < numReactants; j++){
		perm.push_back(j);
	}
	bool successRe = true;
	for(int j=0; j < factorial(numReactants); j++){
		for(int k=0; k < perm.size(); k++){
			if(re1->getReactant(k)->getSpecies() != re2->getReactant(perm.at(k))->getSpecies()){
				successRe = false;
				return false;
			}
		}
		if(successRe) break;
		successRe = true;
		next_permutation(perm.begin(), perm.end());
	}
		
	perm.clear();
	perm.reserve(numModifiers);
	for(int j=0; j < numModifiers; j++){
		perm.push_back(j);
	}
	bool successMod = true;
	for(int j=0; j < factorial(numModifiers); j++){
		for(int k=0; k < perm.size(); k++){
			if(re1->getModifier(k)->getSpecies() != re2->getModifier(perm.at(k))->getSpecies()){
				successMod = false;
				return false;
			}
		}
		if(successMod) break;
		successMod = true;
		next_permutation(perm.begin(), perm.end());
	}
		
	perm.clear();
	perm.reserve(numProducts);
	for(int j=0; j < numProducts; j++){
		perm.push_back(j);
	}
	bool successPr = true;
	for(int j=0; j < factorial(numProducts); j++){
		for(int k=0; k < perm.size(); k++){
			if(re1->getProduct(k)->getSpecies() != re2->getProduct(perm.at(k))->getSpecies()){
				successPr = false;
				return false;
			}
		}
		if(successPr) break;
		successPr = true;
		next_permutation(perm.begin(), perm.end());
	}
	
	if(successRe && successMod && successPr){
		// remove white spaces from kinetic laws
		string formula1 = re1->getKineticLaw()->getFormula();
		string formula2 = re2->getKineticLaw()->getFormula();
		string::iterator it = formula1.begin();
		for(; it != formula1.end(); it++){
			if(*it == ' '){
				formula1.erase(it);
				it--;
			}
		}
		it = formula2.begin();
		for(; it != formula2.end(); it++){
			if(*it == ' '){
				formula2.erase(it);
				it--;
			}
		}
		// substitute parameter names
		int index;
		int len;
		for(int k=0; k < re2->getKineticLaw()->getNumParameters(); k++){
			string id1 = re1->getKineticLaw()->getParameter(k)->getId();
			string id2 = re2->getKineticLaw()->getParameter(k)->getId();
			index = formula1.find(id1);
			len = id1.length();
			while(index != string::npos){
				if(index == ((int)formula1.size()-len)) formula1 = formula1.substr(0, index) + id2;
				else formula1 = formula1.substr(0, index) + id2 + formula1.substr(index+len);
				index = formula1.find(id1,index+len);
			}			
		}
		if(formula1 == formula2) return true;
	}
	
	return false;
}


/**
 * Some simulators (stochastic ones) cannot deal with reversible reactions.
 * This function turns all reversible reactions into two irreversible ones.
 * 
 * Turn all reversible reactions into two reactions, since the gillespie implementation
 * does not support reversible reactions
 * Check that the forward and backward parts of the rate law are divided by '-'
 */
void makeReactionsIrreversible(Model* mod){
	Reaction* re, backRe;
	SpeciesReference* sp;
	ModifierSpeciesReference* msp;
	for(int i=0; i < mod->getNumReactions(); i++){
		re = mod->getReaction(i);
		if(re->getReversible()){
			Reaction* backRe = mod->createReaction();
			backRe->setAnnotation(re->getAnnotation());
			backRe->setId(re->getId() + string("_b"));
			re->setId(re->getId() + string("_f"));
			backRe->setKineticLaw(re->getKineticLaw());
			for(int j=0; j < re->getNumReactants(); j++){
				backRe->addProduct(re->getReactant(j));
			}
			for(int j=0; j < re->getNumProducts(); j++){
				backRe->addReactant(re->getProduct(j));
			}
			for(int j=0; j < re->getNumModifiers(); j++){
				backRe->addModifier(re->getModifier(j));
			}
			const ASTNode* math = re->getKineticLaw()->getMath();
			if(math->getType() != AST_MINUS) THROW_ERROR("Wrong type of formula in kinetic law: reversible reaction cannot be converted into two irreversible ones!");
			if(math->getNumChildren() != 2) THROW_ERROR("Wrong type of formula in kinetic law: reversible reaction cannot be converted into two irreversible ones!");
			ASTNode* forwardRate = math->getLeftChild()->deepCopy();
			ASTNode* backwardRate = math->getRightChild()->deepCopy();
			re->getKineticLaw()->setMath(forwardRate);
			backRe->getKineticLaw()->setMath(backwardRate);
			re->setReversible(false);
			backRe->setReversible(false);
			//cout << re->getKineticLaw()->getFormula() << endl;
			//cout << backRe->getKineticLaw()->getFormula() << endl;
			// Give the reactions the same block id so they will be replaced and deleted together
			int max_block_id = 0;
			int tmp_int;
			for(int j=0; j < mod->getNumReactions(); j++){
				if(isSetAttr(mod->getReaction(j),"block_id")){
					tmp_int = (int) getAttr(mod->getReaction(j),"block_id");
					if(tmp_int > max_block_id) max_block_id = tmp_int;
				}
			}
			if(isSetAttr(re,"block_id")) THROW_ERROR("Trying to set block id where it is already set");
			if(isSetAttr(backRe,"block_id")) THROW_ERROR("Trying to set block id where it is already set");
			setAttr(re,"block_id",max_block_id+1);
			setAttr(backRe,"block_id",max_block_id+1);
		}
	}
}


/**
 * SBML supports stoichiometries >1, but the mutator so far only produces reactions
 * with multiple times the same species and reactants or product, no stoichiometries > 1
 * 
 * Therefore, this functions creates the proper stoichiometries in the model.
 */
void cleanStoichiometries(Model* m){
	Reaction* re;
	SpeciesReference* sr1, *sr2;
	for(int i=0; i < m->getNumReactions(); i++){
		re = m->getReaction(i);
		for(int j=0; j < re->getNumReactants(); j++){
			sr1 = re->getReactant(j);
			for(int k=j+1; k < re->getNumReactants(); k++){
				sr2 = re->getReactant(k);
				if(sr1->getSpecies() == sr2->getSpecies()){
					re->getListOfReactants()->remove(k);
					k--;
					sr1->setStoichiometry(sr1->getStoichiometry() + sr2->getStoichiometry());
					delete sr2;
					sr2 = NULL;
				}
			}
		}
		for(int j=0; j < re->getNumProducts(); j++){
			sr1 = re->getProduct(j);
			for(int k=j+1; k < re->getNumProducts(); k++){
				sr2 = re->getProduct(k);
				if(sr1->getSpecies() == sr2->getSpecies()){
					re->getListOfProducts()->remove(k);
					k--;
					sr1->setStoichiometry(sr1->getStoichiometry() + sr2->getStoichiometry());
					delete sr2;
					sr2 = NULL;
				}
			}
		}
	}
}


string SerializeSBMLModel(Model* m) {
	char* buf;
	SBMLDocument *doc;
	SBMLWriter *w;
	
	// init obj's
	w = new SBMLWriter();
	doc = new SBMLDocument();
	doc->setModel(m);
	
	// serialize model
	buf = w->writeToString(doc);
	
	// deconstruct obj's
	delete doc;
	m->setSBMLDocument(NULL);
	delete w;
	string s(buf);
	
	free(buf);

	return s;
}

Model* DeserializeSBMLModel(string sbmlstring) {
	Model* m;
	SBMLDocument *doc;
	SBMLReader *r;
	
	if(sbmlstring.size() == 0){
		THROW_ERROR("No model string to deserialize!\n");
	}
	
	// read document
	r = new SBMLReader();
	doc = r->readSBMLFromString(sbmlstring.c_str());
	delete r;
	
	m = (Model*) doc->getModel()->clone();
	m->setSBMLDocument(NULL);
	delete doc;
	
	return m;
}

Model* LoadSBMLModel(const char *file_name) {
	Model* m;
	SBMLDocument *doc;
	SBMLReader *r;

	// load document
	r = new SBMLReader();
	doc = r->readSBML(file_name);
	// Check
	if(doc->getNumErrors() > 0){
		doc->printErrors();
		THROW_ERROR("Could not read SBML file!");
	}
	delete r;
	
	// Set SBML format to level 2
	doc->setLevelAndVersion(2,3);

	// and get it out of the doc
	m = (Model*) doc->getModel()->clone();
	m->setSBMLDocument(NULL);
	delete doc;
	
	return m;
}

void SaveSBMLModel(Model* m, const char *file_name) {
	SBMLDocument *doc;
	SBMLWriter *w;
	//string s;
	
	doc = new SBMLDocument();
	doc->setModel(m);
	
	w = new SBMLWriter();
	w->writeSBML(/*(SBMLDocument*)*/ doc, std::string(file_name));
	delete w;

	delete doc;
	m->setSBMLDocument(NULL);
}
	

	/*
	  Format for SBMLevolver annotations in SBase annotation fields is:
	  <annotation>
	      <sbmlevolverData xmlns:\"http://www.minet.uni-jena.de/~thlenser\">
  	          <lower_bound value="0.1"></lower_bound>
		  ...
	      </sbmlevolverData>
	  </annotation>
	 */


#define DONOTMUTATE "<mutate value=\"false\"/>"
#define SBMLEVOLVER_NAMESPACE "<sbmlevolverData xmlns=\"http://www.minet.uni-jena.de/thlenser\">"
#define SBMLEVOLVER_NAMESPACE_END "</sbmlevolverData>"

bool isMutational(SBase* sb){
	if(! sb->isSetAnnotation()) return true;
	string annot = sb->getAnnotationString();
	if(annot.find(DONOTMUTATE) != string::npos){
		return false;
	}
	else return true;
}

void setMutational(SBase* sb, bool mutational){
	if(mutational == false){
		if((! sb->isSetAnnotation() || (sb->getAnnotationString().find(DONOTMUTATE) == string::npos))){
			insertIntoAnnotation(sb, string(DONOTMUTATE));
		}
	}
	else{
		removeFromAnnotation(sb,string(DONOTMUTATE));
	}
}

void insertIntoAnnotation(SBase* sb, string insertme){
	if(sb->isSetAnnotation() && (sb->getAnnotationString().find(SBMLEVOLVER_NAMESPACE) != string::npos)){
		// now we can assume that <annotation> is wrapped all around everything...
		string annot = sb->getAnnotationString();
		int index = annot.find(SBMLEVOLVER_NAMESPACE);
		annot = annot.substr(0,index+strlen(SBMLEVOLVER_NAMESPACE)) + insertme + annot.substr(index+strlen(SBMLEVOLVER_NAMESPACE));
		sb->setAnnotation(annot);
	}
	else{
		if(!sb->isSetAnnotation()){
			string annot = string(string("<annotation>") + string(SBMLEVOLVER_NAMESPACE) + insertme + string(SBMLEVOLVER_NAMESPACE_END) + string("</annotation>"));
			sb->setAnnotation(annot);
		}
		else if(sb->getAnnotationString().find("<annotation>") != string::npos){
			// there is an <annotation> to level element
			string annot = sb->getAnnotationString();
			int index = annot.find("<annotation>");
			annot = annot.substr(0,index+strlen("<annotation>")) + string(SBMLEVOLVER_NAMESPACE) + insertme + string(SBMLEVOLVER_NAMESPACE_END) + annot.substr(index+strlen("<annotation>"));
			sb->setAnnotation(annot);
		}
		else{
			// annotation has no top-level <annotation> wrapper, that's a bug
			cerr << "wrong format of annotation string: " << sb->getAnnotation() << endl;
			exit(-1);
		}
	}
}

void removeFromAnnotation(SBase* sb, string removeme){
	if((sb->isSetAnnotation() && (sb->getAnnotationString().find(removeme) != string::npos))){
		string annot = sb->getAnnotationString();
		int index = annot.find(removeme);
		if(annot.size() > index + removeme.size()) annot = annot.substr(0,index) + annot.substr(index + removeme.size());
		else annot = annot.substr(0,index);
	}
}


/**
 * Functions to get, set and test for attributions in the Annotation tag of 
 * SBase objects. These have the form
 * <attribute value="some_value"/>
 */

void checkAttrName(string attr){
	if((attr != string("lower_bound")) && (attr != string("upper_bound")) && (attr != string("initial_stepsize")) && (attr != string("output_species")) && (attr != string("block_id")) && (attr != string("transport_reaction"))){
		cerr << "Attribute type " << attr << " does not exist" << endl;
		exit(-1);
	}
}

void setAttr(SBase* sb, string attr, double val){
	//	cout << "setAttr " << attr << " : " << val << endl;
	checkAttrName(attr);
	if(isSetAttr(sb,attr)) unsetAttr(sb,attr);
	stringstream strtmp;
	strtmp << "<" << attr << " value=\"" << val << "\"/>";
	//	cout << strtmp.str() << endl;
	insertIntoAnnotation(sb, strtmp.str()); 
}

void setAttr(SBase* sb, string attr, string val){
	checkAttrName(attr);
	if(isSetAttr(sb,attr)) unsetAttr(sb,attr);
	stringstream strtmp;
	strtmp << "<" << attr << " value=\"" << val << "\"/>";
	insertIntoAnnotation(sb, strtmp.str()); 
}

double getAttr(SBase* sb, string attr){
	checkAttrName(attr);
	string annot = sb->getAnnotationString();
	int index = annot.find("<" + attr + " value=\"");
	if(index == string::npos){
		cerr << attr << " was not set" << endl;
		cerr << annot << endl;
		exit(-1);
	}
	int index2 = annot.find("\"/>");
	if(index2 == string::npos){
		cerr << "Error in finding second part of " << attr << endl;
		exit(-1);
	}
	int attr_length_1 = attr.length() + 9;
	string number = annot.substr(index + attr_length_1, index2 - index - attr_length_1);
	return atof(number.c_str());
}

bool isSetAttr(SBase* sb, string attr){
	checkAttrName(attr);
	if(! sb->isSetAnnotation()) return false;
	string annot = sb->getAnnotationString();
	if(annot.find(attr) != string::npos) return true;
	else return false;
}

void unsetAttr(SBase* sb, string attr){
	checkAttrName(attr);
	string annot = sb->getAnnotationString();
	int index1 = annot.find("<" + attr + " value=\"");
	if(index1 == string::npos){
		THROW_ERROR("Trying to unset an attribute that cannot be found!\n");
	}
	int index2 = annot.find("\"/>");
	if(index2 == string::npos){
		THROW_ERROR("Trying to unset an attribute that cannot be fully found!\n");
	}
	int length_of_end_tag = 3;
	index2 += length_of_end_tag;
	if(annot.at(index2) == '\n'){
		index2 += 1;
	}
	if(annot.size() > index2) annot = annot.substr(0,index1) + annot.substr(index2);
	else annot = annot.substr(0,index1);
	sb->setAnnotation(annot);
}


// Print a reaction
void printReaction(Reaction* re){
	cout << "Reaction name: " << re->getId() << endl;
	cout << "Reactants: ";
	for(int i=0; i < re->getNumReactants(); i++) cout << re->getReactant(i)->getSpecies() << " ";
	cout << endl << "Products: ";
	for(int i=0; i < re->getNumProducts(); i++) cout << re->getProduct(i)->getSpecies() << " ";
	cout << endl << "Modifiers: ";
	for(int i=0; i < re->getNumModifiers(); i++) cout << re->getModifier(i)->getSpecies() << " ";
	cout << endl << "KineticLaw: ";
	cout << re->getKineticLaw()->getFormula() << endl;
}
	
	
// Find the compartment of a reaction
// This only works for non-transport reactions!!!
string getCompartmentOfReaction(Model* m, Reaction* re){
	if(isSetAttr(re,"transport_reaction")){
		THROW_ERROR("Trying to determine the compartment of a transport reaction, this should not happen!");
	}
	
	string compartment; 
	if(re->getNumReactants() > 0){
		compartment = m->getSpecies(re->getReactant(0)->getSpecies())->getCompartment();
	}
	else if(re->getNumProducts() > 0){
		compartment = m->getSpecies(re->getProduct(0)->getSpecies())->getCompartment();
	}
	else if(re->getNumModifiers() > 0){
		compartment = m->getSpecies(re->getModifier(0)->getSpecies())->getCompartment();
	}
	else{
		THROW_ERROR("Reaction has no reactants, modifiers or products!");
	}
	return compartment;
}

// Collect all species that are in the given compartment
vector<int> getSpeciesInCompartment(Model* m, string compartment){
	vector<int> speciesInCompartment;
	if(m->getCompartment(compartment) == NULL){
		cerr << "Asked to select species from compartment " << compartment << " which does not exist!" << endl;
		exit(-1);
	}
	for(int u=0; u < m->getNumSpecies(); u++){
		if(m->getSpecies(u)->getCompartment() == compartment){
			speciesInCompartment.push_back(u);
		}
	}

	return speciesInCompartment;
}



} // namespace

